import typing

import numpy as np
import dataclasses
import collections.abc
import scipy.interpolate
from skimage import measure

from spdm.core.field import Field
from spdm.core.expression import Variable
from spdm.geometry.curve import Curve
from spdm.core.geo_object import GeoObject
from spdm.geometry.point import Point
from spdm.utils.logger import deprecated, logger
from spdm.utils.tags import _not_found_

from .optimize import minimize_filter

# import matplotlib.pyplot as plt
# @deprecated
# def find_contours_matplotlib(z: np.ndarray, x: np.ndarray = None, y: np.ndarray = None, /, *args, levels=None, ** kwargs) -> typing.List[typing.List[np.ndarray]]:
#     """
#         args:X: np.ndarray, Y: np.ndarray, Z: np.ndarray
#         TODO: need improvement
#     """
#     fig = plt.figure()
#     contour_set = fig.gca().contour(x, y, z, *args, levels=levels, ** kwargs)
#     return [(contour_set.levels[idx], col.get_segments()) for idx, col in enumerate(contour_set.collections)]


def find_countours_skimage_(
    val: float, z: np.ndarray, x_inter, y_inter
) -> typing.Generator[GeoObject | None, None, None]:
    for c in measure.find_contours(z, val):
        # data = [[x_inter(p[0], p[1], grid=False), y_inter(p[0], p[1], grid=False)] for p in c]
        x = np.asarray(x_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        y = np.asarray(y_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        data = np.stack([x, y], axis=-1)

        if len(data) == 0:
            yield None
        elif data.shape[0] == 1:
            yield Point(*data[0])
        else:
            yield Curve(data)


def find_countours_skimage(vals: list, z: np.ndarray, x: np.ndarray, y: np.ndarray):
    if z.shape == x.shape and z.shape == y.shape:
        pass
    else:
        raise ValueError(f"Array shape does not match! x:{x.shape} , y:{y.shape}, z:{z.shape} ")
    shape = z.shape
    dim0 = np.linspace(0, shape[0] - 1, shape[0])
    dim1 = np.linspace(0, shape[1] - 1, shape[1])
    x_inter = scipy.interpolate.RectBivariateSpline(dim0, dim1, x)
    y_inter = scipy.interpolate.RectBivariateSpline(dim0, dim1, y)

    if not isinstance(vals, (collections.abc.Sequence, np.ndarray)):
        vals = [vals]
    elif isinstance(vals, np.ndarray) and vals.ndim == 0:
        vals = vals.reshape([1])

    for val in vals:
        yield val, find_countours_skimage_(val, z, x_inter, y_inter)

        # count = 0
        # for c in measure.find_contours(z, val):
        #     count += 1
        #     # data = [[x_inter(p[0], p[1], grid=False), y_inter(p[0], p[1], grid=False)] for p in c]
        #     x = np.asarray(x_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        #     y = np.asarray(y_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        #     data = np.stack([x, y], axis=-1)
        #     if len(data) == 0:
        #         yield val, None
        #     elif data.shape[0] == 1:
        #         yield val, Point(*data[0])
        #     else:
        #         yield val, Curve(data)
        # if count == 0:
        #     yield val, None


def find_contours(
    *args, values, **kwargs
) -> typing.Generator[typing.Tuple[float, typing.Generator[GeoObject | None, None, None]], None, None]:
    if len(args) == 3:
        z, x, y = args
    elif len(args) == 1:
        if not isinstance(args[0], Field):
            raise TypeError(f"Wrong type of argument! should be Field, got {type(args[0])}")
        f = args[0]
        xy = f.mesh.points
        x = xy[0]
        y = xy[1]
        z = np.asarray(f)
    else:
        raise ValueError(f"Wrong number of arguments! should be 1 or 3, got {len(args)}")

    yield from find_countours_skimage(values, z, x, y, **kwargs)


@dataclasses.dataclass
class OXPoint:
    r: float
    z: float
    value: float


def find_critical_points(psi: Field) -> typing.Tuple[typing.Sequence[OXPoint], typing.Sequence[OXPoint]]:
    opoints = []

    xpoints = []

    R, Z = psi.mesh.points
    _R = Variable(0, "R")
    Bp2 = (psi.pd(0, 1) ** 2 + psi.pd(1, 0) ** 2) / (_R**2)

    D = psi.pd(2, 0) * psi.pd(0, 2) - psi.pd(1, 1) ** 2

    for r, x_z in minimize_filter(Bp2, R, Z):
        p = OXPoint(r, x_z, psi(r, x_z))

        if D(r, x_z) < 0.0:  # saddle/X-point
            xpoints.append(p)
        else:  # extremum/ O-point
            opoints.append(p)

    Rmid, Zmid = psi.mesh.geometry.bbox.origin + psi.mesh.geometry.bbox.dimensions * 0.5

    opoints.sort(key=lambda x: (x.r - Rmid) ** 2 + (x.z - Zmid) ** 2)

    # TODO:

    o_psi = opoints[0].value
    o_r = opoints[0].r
    o_z = opoints[0].z

    # remove illegal x-points . learn from freegs
    # check psi should be monotonic from o-point to x-point

    x_points = []
    s_points = []
    for xp in xpoints:
        length = 20

        psiline = psi(np.linspace(o_r, xp.r, length), np.linspace(o_z, xp.z, length))

        if len(np.unique(psiline[1:] > psiline[:-1])) != 1:
            s_points.append(xp)
        else:
            x_points.append(xp)

    xpoints = x_points

    xpoints.sort(key=lambda x: (x.value - o_psi) ** 2)

    if len(opoints) == 0 or len(xpoints) == 0:
        raise RuntimeError(f"Can not find O-point or X-point! {opoints} {xpoints}")

    return opoints, xpoints


def find_contours_old(psirz: Field, psi, axis=None) -> typing.Generator[typing.Tuple[float, GeoObject], None, None]:
    """
    if axis is not None:
        only return  closed surface  enclosed axis
        if closed surface does not exists, return None
        number of surface == len(psi)
    else:
        do not guarantee the number of surface == len(psi)
        return all surface ,
    """
    if psi is None or psi is _not_found_:
        psi = []
    elif isinstance(psi, float):
        psi = [psi]

    if axis is None or axis is False:
        for psi_val, surfs in find_contours(psirz, values=psi):
            for surf in surfs:
                if isinstance(surf, GeoObject):
                    surf.set_coordinates("r", "z")
                yield psi_val, surf

    else:
        # x_point = None
        if isinstance(axis, OXPoint):
            o_point = axis
        elif isinstance(axis, (list, tuple)):
            o_point = OXPoint(*axis)
        else:
            raise TypeError(f"Wrong type of argument! should be OXPoint, list or tuple, got {type(axis)}")

        if isinstance(psi, float):
            psi = [psi]

        current_psi = np.nan
        current_count = 0
        for psi_val, surfs in find_contours(psirz, values=psi):
            # 累计相同 level 的 surface个数
            # 如果累计的 surface 个数大于1，说明存在磁岛
            # 如果累计的 surface 个数等于0，说明该 level 对应的 surface 不存在
            # 如果累计的 surface 个数等于1，说明该 level 对应的 surface 存在且唯一

            count = 0
            for surf in surfs:
                count += 1
                if surf is None and np.isclose(psi_val, o_point.value):
                    yield psi_val, Point(o_point.r, o_point.z)
                elif isinstance(surf, Point) and all(np.isclose(surf.points, [o_point.r, o_point.z])):
                    yield psi_val, surf  # raise RuntimeError(f"Can not find surface psi={level}")
                elif isinstance(surf, Curve) and surf.is_closed and surf.enclose(o_point.r, o_point.z):
                    # theta_0 = np.arctan2(x_point.r-o_point.r, x_point.z-o_point.z)
                    # theta = ((np.arctan2(_R-o_point.r, _Z-o_point.z)-theta_0)+2.0*scipy.constants.pi) % (2.0*scipy.constants.pi)
                    # surf = surf.remesh(theta)
                    surf.set_coordinates("r", "z")
                    yield psi_val, surf
                else:
                    count -= 1
            if count <= 0:
                if np.isclose(psi_val, o_point.value):
                    yield psi_val, Point(o_point.r, o_point.z)
                else:
                    # logger.warning(f"{psi_val} {o_point.psi}")
                    yield psi_val, None
            elif current_count > 1:
                raise RuntimeError(f"Something wrong! Get {current_count} closed surfaces for psi={current_psi}")

            # theta = np.arctan2(surf[:, 0]-o_point.r, surf[:, 1]-o_point.z)
            # logger.debug((max(theta)-min(theta))/(2.0*scipy.constants.pi))
            # if 1.0 - (max(theta)-min(theta))/(2.0*scipy.constants.pi) > 2.0/len(theta):  # open or do not contain o-point
            #     current_count -= 1
            #     continue

            # is_closed = False

            # if np.isclose((theta[0]-theta[-1]) % (2.0*scipy.constants.pi), 0.0):
            #     # 封闭曲线
            #     theta = theta[:-1]
            #     surf = surf[:-1]
            #     # is_closed = True
            # else:  # boundary separatrix
            #     if x_point is None:
            #         raise RuntimeError(f"No X-point ")
            #     # logger.warning(f"The magnetic surface average is not well defined on the separatrix!")
            #     xpt = np.asarray([x_point.r, x_point.z], dtype=float)
            #     b = surf[1:]
            #     a = surf[:-1]
            #     d = b-a
            #     d2 = d[:, 0]**2+d[:, 1]**2
            #     p = xpt-a

            #     c = (p[:, 0]*d[:, 0]+p[:, 1]*d[:, 1])/d2
            #     s = (p[:, 0]*d[:, 1]-p[:, 1]*d[:, 0])/d2
            #     idx = np.flatnonzero(np.logical_and(c >= 0, c**2+s**2 < 1))

            #     if len(idx) == 2:

            #         idx0 = idx[0]
            #         idx1 = idx[1]

            #         theta_x = np.arctan2(xpt[0]-o_point.r, xpt[1]-o_point.z)

            #         surf = np.vstack([[xpt], surf[idx0:idx1]])
            #         theta = np.hstack([theta_x, theta[idx0:idx1]])
            #     else:
            #         raise RuntimeError(f"Can not get closed boundary {o_point}, {x_point} {idx} !")

            # # theta must be strictly increased
            # p_min = np.argmin(theta)
            # p_max = np.argmax(theta)

            # if p_min > 0:
            #     if p_min == p_max+1:
            #         theta = np.roll(theta, -p_min)
            #         surf = np.roll(surf, -p_min, axis=0)
            #     elif p_min == p_max-1:
            #         theta = np.flip(np.roll(theta, -p_min-1))
            #         surf = np.flip(np.roll(surf, -p_min-1, axis=0), axis=0)
            #     else:
            #         raise ValueError(f"Can not convert 'u' to be strictly increased!")
            #     theta = np.hstack([theta, [theta[0]+(2.0*scipy.constants.pi)]])
            #     theta = (theta-theta.min())/(theta.max()-theta.min())
            #     surf = np.vstack([surf, surf[:1]])

            # if surf.shape[0] == 0:
            #     logger.warning(f"{level},{o_point.psi},{(max(theta),min(theta))}")

            # elif surf.shape[0] == 1:
            #     yield level, Point(surf[0][0], surf[0][1])
            # else:
            #     yield level, Curve(surf, theta, is_closed=is_closed)



import numpy as np

from spdm.core.expression import derivative
from spdm.core.sp_property import sp_tree

from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_transport import CoreTransport
from fytok.modules.equilibrium import Equilibrium

from .predefined import PredefinedTransport


@sp_tree
class FastAlpha(PredefinedTransport):
    """
    FastAlpha   Model

    Reference:

    [1] Angioni, C., Peeters, A. G., Pereverzev, G. V., Bottino, A., Candy, J., Dux, R., Fable, E., Hein, T., & Waltz, R. E. (2009).
        Gyrokinetic simulations of impurity, He ash and \\alpha particle transport and consequences on ITER transport modelling.
        Nuclear Fusion, 49(5), 055013. https://doi.org/10.1088/0029-5515/49/5/055013
    [2] Waltz, R. E., & Bass, E. M. (2014).
        Prediction of the fusion alpha density profile in ITER from local marginal stability to Alfvén eigenmodes.
        Nuclear Fusion, 54(10), 104006. https://doi.org/10.1088/0029-5515/54/10/104006

    """

    identifier = "alpha"
    code = {"name": "fast_alpha", "description": f" Fast alpha, Angioni's model"}

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D, *args, **kwargs) -> CoreTransport.Model.TimeSlice:
        current: CoreTransport.Model.TimeSlice = super().fetch(profiles_1d, *args, **kwargs)

        eq: Equilibrium.TimeSlice = self.inports["equilibrium/time_slice/current"].fetch()

        R0 = eq.vacuum_toroidal_field.r0

        rho_tor_norm = profiles_1d.rho_tor_norm

        # _x = rho_tor_norm

        # chi = current.profiles_1d.ion["D"].energy.d

        # chi_e = current.profiles_1d.electrons.energy.d

        D = current.profiles_1d.ion["D"].particles.d  # 0.1 * (chi + chi_e)

        Te = profiles_1d.electrons.temperature

        inv_L_Te = -derivative(Te, rho_tor_norm) / Te

        Te_Ea = Te / 3.5e6  # Te/ 3.5MeV

        Ec_Ea = 33.05 * Te_Ea

        r_ped = 0.96  #

        delta = np.heaviside(rho_tor_norm - r_ped, 0.5)

        fast_factor_d = (0.02 + 4.5 * (Te_Ea) + 8.0 * (Te_Ea**2) + 350 * (Te_Ea**3)) * D

        Cs = 1.5 * (1.0 / np.log((Ec_Ea ** (-1.5) + 1) * (Ec_Ea**1.5 + 1)) - 1) * inv_L_Te

        fast_factor_v = fast_factor_d * rho_tor_norm / R0  # Cs * (1 - delta)

        current.profiles_1d.ion = [{"@name": "alpha", "particles": {"d": fast_factor_d, "v": -fast_factor_v}}]

        return current


CoreTransport.Model.register(["fast_alpha"], FastAlpha)



from functools import cached_property

from  ..ontology import edge_sources import _T_edge_sources, _T_edge_sources_source
from spdm.utils.tags import _undefined_
from spdm.core.Dict import Dict
from spdm.core.htree import List
from spdm.core.sp_property import sp_property, SpTree


class EdgeSources(_T_edge_sources):

    Source = _T_edge_sources_source

    @cached_property
    def source_combiner(self) -> Source:
        return self.source.combine({
            "identifier": {"name": "total", "index": 1,
                           "description": "Total source; combines all sources"},
            "code": {"name": _undefined_}
        })

    def update(self, *args,   **kwargs) -> float:
        if "source_combiner" in self.__dict__:
            del self.__dict__["source_combiner"]
        return sum([src.refresh(*args,   **kwargs) for src in self.source])



import getpass
import os
import datetime
from spdm.core.sp_property import sp_tree, sp_property

from ..ontology import dataset_fair, GLOBAL_ONTOLOGY


@sp_tree
class DataDescription:
    device: str

    shot: int

    run: int = sp_property(default_value=0)

    summary: str = sp_property(default_value="")

    def __str__(self) -> str:
        return f"{self.device.upper()} #{self.shot}/{self.run}"

    @sp_property
    def tag(self) -> str:
        return f"{self.device.lower()}_{self.shot}_{self.run}"


@sp_tree
class DatasetFAIR(dataset_fair._T_dataset_fair):
    ontology: str = GLOBAL_ONTOLOGY

    description: DataDescription

    @sp_property
    def creator(self) -> str:
        return getpass.getuser().capitalize()

    @sp_property
    def create_time(self) -> str:
        return datetime.datetime.now().isoformat()

    @sp_property
    def site(self) -> str:
        return os.uname().nodename

    def __str__(self) -> str:
        return f""" 
    Device: {self.description.device.upper()}, Shot: {self.description.shot}, Run: {self.description.run}, 
    Run by {self.creator} on {self.site} at {self.create_time}, base on ontology \"{self.ontology}\"
"""



from __future__ import annotations


from spdm.core.geo_object import GeoObject

from ..ontology import pellets


class Pellets(pellets._T_pellets):
    def __view__(self, view="RZ", **kwargs):
        return {} 



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

import os
import typing

from ..utils.logger import logger


GLOBAL_ONTOLOGY = os.environ.get("FY_ONTOLOGY", f"imas/3")


try:
    from .imas_lastest.__version__ import __version__ as imas_version

    from .imas_lastest import (
        dataset_fair,
        summary,
        pulse_schedule,
        equilibrium,
        core_profiles,
        core_sources,
        ic_antennas,
        interferometer,
        lh_antennas,
        magnetics,
        nbi,
        pellets,
        core_transport,
        pf_active,
        tf,
        transport_solver_numerics,
        utilities,
        ec_launchers,
        amns_data,
        wall,
        waves,
    )

    __all__ = [
        "dataset_fair",
        "summary",
        "pulse_schedule",
        "equilibrium",
        "core_profiles",
        "core_sources",
        "ec_launchers",
        "ic_antennas",
        "interferometer",
        "lh_antennas",
        "magnetics",
        "nbi",
        "pellets",
        "amns_data",
        "core_transport",
        "wall",
        "pf_active",
        "tf",
        "transport_solver_numerics",
        "utilities",
        "waves",
    ]


except ModuleNotFoundError as error:
    logger.verbose(f"Failed to import IMAS ontology: {error}")

    logger.info(f"Failed to import IMAS wrapper. Using dummy ontology as fallback.")

    from spdm.core.sp_property import PropertyTree

    imas_version = "None"

    class DummyModule:
        def __init__(self, name):
            self._module = name

        def __str__(self) -> str:
            return f"<dummy_module '{__package__}.dummy.{self._module}'>"

        def __getattr__(self, __name: str) -> typing.Type[PropertyTree]:
            cls = type(__name, (PropertyTree,), {})
            cls.__module__ = f"{__package__}.dummy.{self._module}"
            return cls

    def __getattr__(key: str) -> DummyModule:
        return DummyModule(key)

else:
    if GLOBAL_ONTOLOGY != f"imas/{imas_version[1:].split('_')[0]}":
        raise RuntimeError(
            f"Global ontology {GLOBAL_ONTOLOGY} is not compatible with IMAS version {imas_version[1:].split('.')[0]}"
        )



import pathlib
from fytok.tokamak import Tokamak

WORKSPACE = "/home/salmon/workspace"  # "/ssd01/salmon_work/workspace/"
output_path = pathlib.Path(f"{WORKSPACE}/output/")

tok = Tokamak(
    "file+geqdsk:///home/salmon/workspace/fytok_tutorial/tutorial/data/g900003.00230_ITER_15MA_eqdsk16HR.txt",
    device="east",
    equilibrium={"code": {"name": "freegs"}},
)

tok.refresh()


from __future__ import annotations

import typing

from spdm.core.geo_object import GeoObject
from spdm.geometry.line import Line
from spdm.geometry.polygon import Rectangle

from ..ontology import nbi


def draw_nbi_unit(unit: nbi._T_nbi_unit, name: str):
    geo = None
    if unit.source.geometry_type == 3:
        geo = [Line(), Rectangle(name=unit.name)]

    else:
        pass

    return geo


class NBI(nbi._T_nbi):
    def __view__(self, view_point="RZ", **kwargs) -> GeoObject:
        geo = {}
        styles = {}

        match view_point.lower():
            case "top":
                geo["unit"] = [draw_nbi_unit(unit) for unit in self.unit]

        return geo




import numpy as np
from scipy import constants
from spdm.utils.tags import _not_found_
from spdm.core.Dict import Dict
from spdm.core.entry import as_entry
from spdm.core.expression import Expression 
from spdm.core.htree import List
from spdm.core.Node import Node
from spdm.core.sp_property import sp_property, SpTree

from .Atoms import atoms
from .MagneticCoordSystem import RadialGrid


class Species(Dict[Node]):

    Element = SpeciesElement

    label: str = sp_property()
    """String identifying ion (e.g. H+, D+, T+, He+2, C+, ...) {dynamic}    """

    @property
    def grid(self) -> RadialGrid:
        return self._parent.grid

     

    multiple_states_flag: int = sp_property(default=0)

    @sp_property
    def element(self, value) -> List[Element]:
        if value is None:
            value = as_entry(atoms).get(f"{self.label}/element")
        return value

    @sp_property
    def a(self, value) -> float:
        """Mass of ion {dynamic} [Atomic Mass Unit]"""
        if value is None:
            value = sum([a.a * a.atoms_n for a in self.element])
        return value

    @sp_property
    def z(self, value) -> float:
        if value is None:
            value = as_entry(atoms).get(f"{self.label}/z")
        return value



import os
import pathlib

from fytok.tokamak import Tokamak
from spdm.view.sp_view import display

WORKSPACE = "/home/salmon/workspace"  # "/ssd01/salmon_work/workspace/"

os.environ["SP_DATA_MAPPING_PATH"] = f"{WORKSPACE}/fytok_data/mapping"

if __name__ == "__main__":

    output_path = pathlib.Path(f"{WORKSPACE}/output")

    # entry = open_entry(f"EAST+file+iterdb://{WORKSPACE}/gacode/neo/tools/input/profile_data/iterdb141459.03890")

    # logger.debug(entry.get("equilibrium/time_slice[0]/vacuum_toroidal_field/r0"))

    # tok = Tokamak(f"D3D+file+iterdb://{WORKSPACE}/gacode/neo/tools/input/profile_data/iterdb141459.03890",
    #               equilibrium={"code": {"name": "eq_analyze"},
    #                            "$entry": [f"file+geqdsk://{WORKSPACE}/gacode/neo/tools/input/profile_data/g141459.03890#equilibrium"]
    #                            }
    #               )
    tok = Tokamak(device="D3D",
                  entry=[
                      f"file+iterdb://{WORKSPACE}/gacode/neo/tools/input/profile_data/iterdb141459.03890",
                      f"file+geqdsk://{WORKSPACE}/gacode/neo/tools/input/profile_data/g141459.03890"
                  ],
                  equilibrium={"code": {"name": "eq_analyze"}})

    display(tok, title=f"{tok.device.upper()} RZ   View ", output=output_path/"d3d_rz.svg")
    # display(tok, title=f"{tok.device.upper()} Top  View ", output=output_path/"d3d_top.svg", view_point="TOP")



import typing
import pathlib
import numpy as np
from scipy import interpolate
from spdm.core.entry import Entry, as_entry
from spdm.core.file import File, FileEntry
from spdm.core.field import Field
from spdm.core.function import Function

from spdm.utils.logger import logger


def sp_read_geqdsk(file):
    """
    :param file: input file / file path
    :return: profile object
    """

    description = file.read(48)
    idum = int(file.read(4))
    nw = int(file.read(4))
    nh = int(file.read(4))
    file.readline()

    rdim = float(file.read(16))
    zdim = float(file.read(16))
    rcentr = float(file.read(16))
    rleft = float(file.read(16))
    zmid = float(file.read(16))
    file.readline()
    rmaxis = float(file.read(16))
    zmaxis = float(file.read(16))
    simag = float(file.read(16))
    sibry = float(file.read(16))
    bcentr = float(file.read(16))
    file.readline()
    current = float(file.read(16))
    simag = float(file.read(16))
    xdum = float(file.read(16))
    rmaxis = float(file.read(16))
    xdum = float(file.read(16))
    file.readline()

    zmaxis = float(file.read(16))
    xdum = float(file.read(16))
    sibry = float(file.read(16))
    xdum = float(file.read(16))
    xdum = float(file.read(16))
    file.readline()

    def _read_data(count, width=16):
        data = []
        if count == 0:
            return data

        for n in range(count):
            d = file.read(width)

            try:
                v = float(d)
            except Exception as error:
                raise RuntimeError(f"Error reading data {n} {count} {data[-4:]} '{d}'") from error
            data.append(v)
            if n >= count - 1 or ((n + 1) % 5 == 0):
                file.readline()
        data = np.asarray(data)
        return data

    #
    fpol = _read_data(nw)
    pres = _read_data(nw)
    ffprim = _read_data(nw)
    pprim = _read_data(nw)

    psirz = _read_data(nw * nh).reshape([nh, nw])

    qpsi = _read_data(nw)

    # try:
    nbbs = int(file.read(5))
    nlimitr = int(file.read(5))
    file.readline()

    bbsrz = _read_data(nbbs * 2).reshape([nbbs, 2]) if nbbs > 0 else None

    limrz = _read_data(nlimitr * 2).reshape([nlimitr, 2]) if nlimitr > 0 else None
    # except Exception as error:

    #     nbbs = 0
    #     limitr = 0
    #     bbsrz = None
    #     limrz = None

    data = {
        "description": description,
        # "idum": idum,
        "nw": nw,
        "nh": nh,
        "rdim": rdim,
        "zdim": zdim,
        "rcentr": rcentr,
        "rleft": rleft,
        "zmid": zmid,
        "rmaxis": rmaxis,
        "zmaxis": zmaxis,
        "simag": simag,
        "sibry": sibry,
        "bcentr": bcentr,
        "current": current,
        # "simag": simag,
        # "rmaxis": rmaxis,
        # "zmaxis": zmaxis,
        # "sibry": sibry,
        "fpol": fpol,
        "pres": pres,
        "ffprim": ffprim,
        "pprim": pprim,
        "psirz": psirz,
        "qpsi": qpsi,
        "bbsrz": bbsrz,
        "limrz": limrz,
    }
    return data


def sp_write_geqdsk(p, file):
    """
    :param profile: object

    :param file: file path / file
    :return:
    """
    # logger.debug(p)
    nw = p["nw"]
    nh = p["nh"]

    file.write("%48s%4i%4i%4i\n" % (p.get("description", "NO DESCRIPTION"), 3, nw, nh))
    file.write("%16.8e%16.8e%16.8e%16.8e%16.8e\n" % (p["rdim"], p["zdim"], p["rcentr"], p["rleft"], p["zmid"]))
    file.write("%16.8e%16.8e%16.8e%16.8e%16.8e\n" % (p["rmaxis"], p["zmaxis"], p["simag"], p["sibry"], p["bcentr"]))
    file.write("%16.8e%16.8e%16.8e%16.8e%16.8e\n" % (p["current"], p["simag"], 0, p["rmaxis"], 0))
    file.write("%16.8e%16.8e%16.8e%16.8e%16.8e\n" % (p["zmaxis"], 0, p["sibry"], 0, 0))

    def _write_data(d):
        if not isinstance(d, np.ndarray):
            logger.debug(d)
        count = len(d)
        for n in range(count):
            file.write("%16.8e" % d[n])
            if (n == count - 1) or ((n + 1) % 5 == 0):
                file.write("\n")
            # else:
            #     file.write(" ")

    _write_data(p["fpol"])
    _write_data(p["pres"])
    _write_data(p["ffprim"])
    _write_data(p["pprim"])
    _write_data(p["psirz"].reshape([nw * nh]))
    _write_data(p["qpsi"])

    bbsrz = p.get("bbsrz", np.zeros([0, 2]))
    limrz = p.get("limrz", np.zeros([0, 2]))

    file.write("%5i%5i\n" % (bbsrz.shape[0], limrz.shape[0]))

    _write_data(bbsrz.reshape([bbsrz.size]))
    _write_data(limrz.reshape([limrz.size]))

    return


def sp_to_geqdsk(d, description: str | None = None, time_slice=0, **kwargs) -> dict:
    entry: Entry = as_entry(d)

    geqdsk: dict = {"description": description or entry.get("description", "NOTHING TO SAY")}

    limiter_r = entry.get("wall/description_2d/0/limiter/unit/0/outline/r", None)
    limiter_z = entry.get("wall/description_2d/0/limiter/unit/0/outline/z", None)

    if isinstance(limiter_r, np.ndarray):
        geqdsk["limrz"] = np.append(
            limiter_r.reshape([1, limiter_r.size]),
            limiter_z.reshape([1, limiter_z.size]),
            axis=0,
        ).transpose()

    eq = entry.child(f"equilibrium/time_slice/{time_slice}")

    # rdim = 0.0
    # zdim = 0.0
    geqdsk["rcentr"] = eq.get("boundary/geometric_axis/r")
    # rleft = 0.0
    geqdsk["zmid"] = zmid = eq.get("boundary/geometric_axis/z")

    geqdsk["rmaxis"] = eq.get("global_quantities/magnetic_axis/r")
    geqdsk["zmaxis"] = eq.get("global_quantities/magnetic_axis/z")
    geqdsk["simag"] = eq.get("global_quantities/psi_axis")
    geqdsk["sibry"] = eq.get("global_quantities/psi_boundary")
    geqdsk["bcentr"] = eq.get("vacuum_toroidal_field/b0")
    geqdsk["current"] = eq.get("global_quantities/ip")

    # boundary

    rbbs = eq.get("boundary/outline/r", np.zeros([0]))
    zbbs = eq.get("boundary/outline/z", np.zeros([0]))

    geqdsk["bbsrz"] = np.append(rbbs.reshape([1, rbbs.size]), zbbs.reshape([1, rbbs.size]), axis=0).transpose()

    # psi

    psirz = eq.get("profiles_2d/psi")

    if eq.get("profiles_2d/grid_type/index", None) != 1:
        raise NotImplementedError(f"TODO: {eq.get('profiles_2d/grid_type/index', None)}")

    dim1 = eq.get("profiles_2d/grid/dim1")
    dim2 = eq.get("profiles_2d/grid/dim2")

    nw = dim1.size
    nh = dim2.size

    geqdsk["rleft"] = dim1.min()
    geqdsk["rmid"] = rmid = 0.5 * (dim1.max() + dim1.min())
    geqdsk["rdim"] = rdim = dim1.max() - dim1.min()
    geqdsk["zdim"] = zdim = dim2.max() - dim2.min()

    # coord_r = np.append(coord_r[:, :], coord_r[:, 0].reshape(coord_r.shape[0], 1), axis=1)
    # coord_z = np.append(coord_z[:, :], coord_z[:, 0].reshape(coord_z.shape[0], 1), axis=1)

    # points = np.append(coord_r.reshape([coord_r.size, 1]), coord_z.reshape([coord_z.size, 1]), axis=1)

    if not isinstance(psirz, np.ndarray):
        psirz = psirz.__array__()

    geqdsk["psirz"] = psirz.T

    geqdsk["nw"] = nw

    geqdsk["nh"] = nh

    # psi = np.append(psi[:, :], psi[:, 0].reshape(psi.shape[0], 1), axis=1)
    # values = psi[:coord_r.shape[0], :coord_r.shape[1]].reshape(points.shape[0])
    # psirz = interpolate.griddata(points, values, (grid_r, grid_z), method='cubic').transpose()

    # profile

    psi = eq.get("profiles_1d/psi", None)

    psi_axis = eq.get("global_quantities/psi_axis")
    psi_boundary = eq.get("global_quantities/psi_boundary")

    psi_ = np.linspace(psi_axis, psi_boundary, nw)
    nan_array = np.full(nw, np.nan)

    def _profile(f, psi_s):
        if f is None or psi_s is None:
            return nan_array
        else:
            if psi is None:
                length = len(f)
                x = np.linspace(psi_axis, psi_boundary, length)
            else:
                x = psi
            try:
                res = Function(x, f)(psi_s)
            except Exception:
                return nan_array
            else:
                return res

    geqdsk["fpol"] = _profile(eq.get("profiles_1d/f", None), psi_)
    geqdsk["pres"] = _profile(eq.get("profiles_1d/pressure", None), psi_)
    geqdsk["ffprim"] = _profile(eq.get("profiles_1d/f_df_dpsi", None), psi_)
    geqdsk["pprim"] = _profile(eq.get("profiles_1d/dpressure_dpsi", None), psi_)
    geqdsk["qpsi"] = _profile(eq.get("profiles_1d/q", None), psi_)

    return geqdsk

    # return Entry({
    #     "nw": nw,
    #     "nh": nh,
    #     "rdim": rdim,
    #     "zdim": zdim,
    #     "rcentr": rcentr,
    #     "rleft": rleft,
    #     "zmid": zmid,
    #     "rmaxis": rmaxis,
    #     "zmaxis": zmaxis,
    #     "simag": simag,
    #     "sibry": sibry,
    #     "bcentr": bcentr,
    #     "current": current,
    #     "bbsrz": bbsrz,
    #     "psirz": psirz,
    #     "fpol": fpol,
    #     "pres": pres,
    #     "ffprim": ffprim,
    #     "pprim": pprim,
    #     "qpsi": qpsi,
    #     "limrz": limrz

    # })


def sp_from_geqdsk(geqdsk: dict, eq: typing.Optional[Entry] = None) -> Entry:
    """Converts a GEQDSK file to an IMAS equilibrium entry.
    @TODO:
        - convert to COCOS 11 !!!
    """

    if eq is None:
        eq = Entry({})

    r0 = geqdsk["rcentr"]
    b0 = geqdsk["bcentr"]
    Ip = geqdsk["current"]
    psi_axis = geqdsk["simag"]
    psi_boundary = geqdsk["sibry"]
    q = geqdsk["qpsi"]

    s_Bp = np.sign(b0)
    s_Ip = np.sign(Ip)
    s_rtp = np.mean(np.sign(q)) / (s_Bp * s_Ip)
    assert np.sign(psi_boundary - psi_axis) == s_Ip

    e_Bp_TWOPI = 1.0
    limrz = geqdsk.get("limrz", None)
    if isinstance(limrz, np.ndarray):
        eq["wall"] = {"description_2d": [{"limiter": {"unit": [{"outline": {"r": limrz[:, 0], "z": limrz[:, 1]}}]}}]}

    eq["equilibrium/time"] = [0.0]
    eq["equilibrium/vacuum_toroidal_field/r0"] = r0
    eq["equilibrium/vacuum_toroidal_field/b0"] = [b0]

    # rleft = 0.0

    # eq["global_quantities.magnetic_axis.b_field_tor"] = geqdsk["bcentr"]
    nw = geqdsk["nw"]
    nh = geqdsk["nh"]
    rmin = geqdsk["rleft"]
    rmax = geqdsk["rleft"] + geqdsk["rdim"]
    zmin = geqdsk["zmid"] - geqdsk["zdim"] / 2
    zmax = geqdsk["zmid"] + geqdsk["zdim"] / 2

    psirz = geqdsk["psirz"]

    if psirz.shape == (nh, nw):
        psirz = psirz.T
        # logger.warning(f"Transposing psirz from {(nh, nw)} to {(nw,nh)}")

    if psirz.shape != (nw, nh):
        raise ValueError(f"Invalid shape for psirz: {psirz.shape}!={(nw, nh)}")

    eq["equilibrium/time_slice"] = [
        {
            "time": 0.0,
            "vacuum_toroidal_field": {"r0": r0, "b0": b0},
            "global_quantities": {
                "magnetic_axis": {
                    "r": geqdsk["rmaxis"],
                    "z": geqdsk["zmaxis"],
                },
                "psi_axis": psi_axis,
                "psi_boundary": psi_boundary,
                "ip": Ip,
            },
            "boundary": {  # Boundary
                "outline": {
                    "r": geqdsk["bbsrz"][:, 0],
                    "z": geqdsk["bbsrz"][:, 1],
                },
                "geometric_axis": {
                    "r": geqdsk["rcentr"],
                    "z": geqdsk["zmid"],
                },
            },
            "profiles_1d": {  # Profiles 1D
                "psi_norm": np.linspace(0.0, 1.0, nw),
                "psi": np.linspace(psi_axis, psi_boundary, nw),
                "f": geqdsk["fpol"],
                "f_df_dpsi": geqdsk["ffprim"],
                "pressure": geqdsk["pres"],
                "dpressure_dpsi": geqdsk["pprim"],
                "q": geqdsk["qpsi"],
            },
            "profiles_2d": {  # profiles 2D
                "type": "total",  # total field
                "grid_type": {"name": "rectangular"},
                "grid": {
                    "dim1": np.linspace(rmin, rmax, nw),
                    "dim2": np.linspace(zmin, zmax, nh),
                    "@type": "rectangular",
                },
                "psi": psirz,
            },
        }
    ]

    return eq


@File.register(["gfile", "geqdsk"])
class GEQdskFile(File):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self._fid = open(pathlib.Path(self.url.path).expanduser().resolve(), mode=self.mode_str)
        except OSError as error:
            raise FileExistsError(f"Can not open file {self.url}! {error}")
        # else:
        #     logger.debug(f"Open File mode={self.mode}  {self.url} ")

    def __del__(self):
        if getattr(self, "_fid", None) is not None:
            self._fid.close()
            self._fid = None

    # def flush(self, *args, **kwargs):
    #     if self.mode & File.Mode.write:
    #         self.save(self.path)

    @property
    def entry(self) -> Entry:
        if self.mode == File.Mode.read:
            return self.read()
        else:
            return FileEntry({}, file=self)

    def read(self, lazy=False) -> Entry:
        return sp_from_geqdsk(sp_read_geqdsk(self._fid))

    def write(self, d, *args, **kwargs):
        geqdsk = sp_to_geqdsk(d, *args, **kwargs)
        sp_write_geqdsk(geqdsk, self._fid)
        self._fid.flush()



from __future__ import annotations

import collections
import collections.abc
import functools
import typing
from enum import Enum

import numpy as np
import scipy.constants
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.equilibrium import Equilibrium
from fytok.modules.magnetics import Magnetics
from fytok.modules.pf_active import PFActive
from fytok.modules.wall import Wall
from fytok.modules.utilities import *

from spdm.core.time_series import TimeSlice
from spdm.core.field import Field
from spdm.core.mesh import Mesh

from spdm.utils.constants import *
from spdm.utils.logger import logger
from spdm.numlib.numeric import bitwise_and, squeeze
from spdm.utils.tags import _not_found_
from spdm.utils.typing import ArrayLike, ArrayType, NumericType, array_type, as_array, as_scalar, is_array, scalar_type

from fytok.plugins.equilibrium.fy_eq import FyEqAnalyze

try:
    import freegs
    import freegs.boundary
except ModuleNotFoundError as error:
    raise ModuleNotFoundError(f"Can not find module 'freegs'!") from error


@sp_tree
class EquilibriumFreeGS(FyEqAnalyze):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._machine: freegs.machine.Machine = None

        self._eq_solver: freegs.Equilibrium = None

    def _setup_machine(self, **kwargs) -> freegs.machine.Machine:
        """update the description of machine
        TODO:
            - check if wall and pf_active is changed
            - update current of coils

        return True if machine is updated
        """
        logger.debug(f"Setup machine")

        pf_active: PFActive = self.inputs.get_source("pf_active")

        eq_coils = []

        for coil in pf_active.coil:
            rect = coil.element[0].geometry.rectangle

            turns = int(coil.element[0].turns_with_sign)
            t_coil = freegs.machine.Coil(rect.r + rect.width / 2, rect.z + rect.height / 2, turns=turns)
            eq_coils.append((coil.name, t_coil))

        wall: Wall = self.inputs.get_source("wall")

        eq_wall = freegs.machine.Wall(
            wall.description_2d[0].limiter.unit[0].outline.r,
            wall.description_2d[0].limiter.unit[0].outline.z,
        )

        magnetics: Magnetics = self.inputs.get_source("magnetics")

        eq_sensors = []

        for b_prob in magnetics.b_field_pol_probe:
            eq_sensors.append(
                freegs.machine.PoloidalFieldSensor(
                    b_prob.position.r,
                    b_prob.position.z,
                    b_prob.position.phi or 0,
                    b_prob.name,
                )
            )

        for flux in magnetics.flux_loop:
            eq_sensors.append(
                freegs.machine.FluxLoopSensor(
                    flux.position[0].r,
                    flux.position[0].z,
                    flux.name,
                )
            )

        for rogowski_coil in magnetics.rogowski_coil:
            eq_sensors.append(
                freegs.machine.RogowskiSensor(
                    rogowski_coil.position.r,
                    rogowski_coil.position.z,
                    rogowski_coil.position.phi or 0,
                    name=b_prob.name,
                )
            )

        self._machine = freegs.machine.Machine(coils=eq_coils, wall=eq_wall, sensors=eq_sensors)

        logger.info(f"Setup machine description: wall={wall.description_2d[0].type}")

        return self._machine

    def preprocess(self, *args, **kwargs):
        super().preprocess(*args, **kwargs)

        current = self.time_slice.current

        if self._eq_solver is None:
            machine = self._setup_machine()

            logger.debug(f"Setup freegs solver")

            boundary_type = self.code.parameters.get("boundary", "free")

            nx, ny = self.code.parameters.get("dims", (128, 128))

            box = self.code.parameters.get("box", None)

            if box is not None:
                Rmin, Zmin, Rmax, Zmax = box
            else:
                wall: Wall = self.inputs.get_source("wall")  # type:ignore
                description_2d = wall.description_2d[0]
                rdim = self.time_slice.current.profiles_2d.grid.dim1
                zdim = self.time_slice.current.profiles_2d.grid.dim2
                # logger.debug(f"Setup freegs solver: rdim={rdim}, zdim={zdim}")
                r: array_type = description_2d.limiter.unit[0].outline.r
                z: array_type = description_2d.limiter.unit[0].outline.z

                Rmin = r.min()
                Rmax = r.max()
                Zmin = z.min()
                Zmax = z.max()

            if boundary_type == "fixed":
                boundary = freegs.boundary.fixedBoundary
            else:
                boundary = freegs.boundary.freeBoundaryHagenow

            # boundary = freegs.boundary.fixedBoundary

            logger.debug(f"Setup freegs solver: boundary={boundary}")

            self._eq_solver = freegs.Equilibrium(
                tokamak=machine,
                # Rmin=Rmin,
                # Rmax=Rmax,
                # Zmin=Zmin,
                # Zmax=Zmax,
                Rmin=min(rdim),
                Rmax=max(rdim),
                Zmin=min(zdim),
                Zmax=max(zdim),
                nx=len(rdim),
                ny=len(zdim),
                boundary=boundary,
            )

            logger.info(f"Using {boundary_type} boundary")

        logger.debug(f"Setup profiles for equilibrium solver")

        ### 从输运算得的core_profiles作为约束
        core_profiles: CoreProfiles = self.inputs.get_source("core_profiles", _not_found_)  # type:ignore

        if core_profiles is not _not_found_:
            core_profiles_1d = core_profiles.time_slice.current.profiles_1d

            R0 = core_profiles.time_slice.current.vacuum_toroidal_field.r0
            B0 = core_profiles.time_slice.current.vacuum_toroidal_field.b0
            fvac: float = R0 * B0  # type:ignore
            psi_norm = core_profiles_1d.grid.psi_norm
        else:
            R0 = kwargs.pop("R0", 1.7)
            B0 = kwargs.pop("B0", 3.0)
            fvac = R0 * B0
            psi_norm = self.code.parameters.get("psi_norm", np.linspace(0, 1.0, 128))

        self._freegs_profiles = None

        ### 用IP ，BETAP，PAXIS作为约束
        #  Poloidal beta
        profiles = {}

        betap = profiles.get("betap", -1.0) or current.global_quantities.beta_pol

        # Plasma current [Amps]
        Ip = profiles.get("ip", None) or current.global_quantities.ip

        B0 = current.vacuum_toroidal_field.b0

        fvac = np.abs(B0 * current.vacuum_toroidal_field.r0)

        betap = -1.7996433287982363 if betap is None else betap

        # logger.debug(f"BETAP = {betap}, IP = {Ip},B0 = {B0}, fvac = {fvac}")

        ## 如果IP betap非空，则用IP betap作为约束

        if Ip is not None and betap is not None:
            self._freegs_profiles = freegs.jtor.ConstrainBetapIp(self._eq_solver, betap, Ip, fvac)

            logger.info(
                f"""Create Profile: Constrain poloidal Beta and plasma current"
                    Betap                      ={betap} [-],
                    Plasma current Ip          ={Ip} [Amps],
                    fvac                       ={fvac} [T.m]"""
            )

        # if "Ip" in kwargs and "beta_p" in kwargs:
        #     Ip = self.code.parameters.Ip

        #     beta_p = self.code.parameters.beta_p

        #     self._freegs_profiles = freegs.jtor.ConstrainBetapIp(self._eq_solver, beta_p, Ip, fvac)

        #     logger.info(
        #         f"""Create Profile: Constrain poloidal Beta and plasma current
        #                 Betap                      = {beta_p} [-],
        #                 Plasma current Ip          = {Ip} [Amp],
        #                 R0*B0                      = {fvac} [T.m]
        #                 """
        #     )
        # elif "pressure_axis" in kwargs and "Ip" in kwargs:
        #     Ip = self.code.parameters.Ip

        #     pressure_axis = self.code.parameters.pressure_axis

        #     self._freegs_profiles = freegs.jtor.ConstrainPaxisIp(self._eq_solver, pressure_axis, Ip, fvac, Raxis=R0)

        #     logger.info(
        #         f"""Create Profile: Constrain pressure on axis and plasma current
        #                 Plasma pressure on axis    = {pressure_axis} [Pascals],
        #                 Plasma current Ip          = {Ip} [Amp ],
        #                 fvac                       = {fvac} [T.m],
        #                 Raxis                      = {R0} [m]
        #                 """
        #     )
        # elif core_profiles is not _not_found_:
        #     core_profiles_1d = core_profiles.time_slice.current.profiles_1d

        #     pprime = core_profiles_1d.pprime(psi_norm)
        #     ffprime = core_profiles_1d.ffprime(psi_norm)

        #     self._freegs_profiles = freegs.jtor.ProfilesPprimeFfprime(pprime, ffprime, fvac)

        #     logger.info("Create Profile: Specified profile functions p'(psi), ff'(psi)")
        # else:
        #     raise RuntimeError(f"Can not create freegs profile!")

        # self._freegs_constraints = None

        # if boundary_type == "fixed":
        #     freegs_constraints = None

        # psivals = kwargs.pop("lcfs", False)

        # psi_bndry = None

        # if psivals is True:
        #     boundary_outline_r = current_time_slice.boundary.outline.r
        #     boundary_outline_z = current_time_slice.boundary.outline.z
        #     boundary_psi = np.full_like(boundary_outline_r, current_time_slice.boundary.psi)

        #     psivals = np.vstack([boundary_outline_r, boundary_outline_z, boundary_psi]).T

        self.psi_bndry = current.boundary.psi

        logger.debug(f"psi_bndry = {self.psi_bndry}")

        #     logger.info(f"Using fixed lcfs")

        # xpoints =
        # if xpoints is True:
        #     xpoints = [(x.r, x.z) for x in current_time_slice.boundary.x_point]
        #     logger.info(f"Using xpoints: {xpoints}")
        # elif isinstance(xpoints, typing.Sequence):
        #     logger.info(f"Using xpoints: {xpoints}")

        # isoflux = kwargs.pop("isoflux", False)
        lfcs_r = current.boundary.outline.r
        lfcs_z = current.boundary.outline.z
        boundary_psi = current.boundary.psi

        # logger.debug(f"lfcs_r = {lfcs_r}, lfcs_z = {lfcs_z}")
        xpoints = []
        isoflux = []
        psivals = [(R, Z, 0.0) for R, Z in zip(lfcs_r, lfcs_z)]

        self._freegs_constraints = freegs.control.constrain(xpoints=xpoints, isoflux=isoflux, psivals=psivals)

        # try:
        #     self._freegs_constraints = freegs.control.constrain(
        #         psivals=kwargs.pop("psivals", []),
        #         isoflux=kwargs.pop("isoflux", []),
        #         xpoints=kwargs.pop("xpoints", []),
        #     )
        # except Exception as error:
        #     raise RuntimeError(f"Can not create freegs constraints!") from error

    def execute(self, current: Equilibrium.TimeSlice, *previous: Equilibrium.TimeSlice):
        """update the last time slice, base on profiles_2d[-1].psi, and core_profiles_1d, wall, pf_active"""
        super().execute(current, *previous)

        rtol = self.code.parameters.get("tolerance", 0.1)

        try:
            logger.info("Solve G-S equation START")
            freegs.solve(
                self._eq_solver,
                self._freegs_profiles,
                constrain=self._freegs_constraints,
                psi_bndry=self.psi_bndry,
                show=True,
                rtol=1.0e-4,
            )
        except Exception as error:
            raise RuntimeError(f"Solve G-S equation failed [{self.__class__.__name__}]!") from error
        else:
            logger.info(f"Solve G-S equation Done")

        print("Plasma current: %e Amps" % (self._eq_solver.plasmaCurrent()))
        print("Plasma pressure on axis: %e Pascals" % (self._eq_solver.pressure(0.0)))
        print("Poloidal beta: %e" % (self._eq_solver.poloidalBeta()))

        # print("Plasma psiRZ: %e psi" % (eq.psi()))
        # psiRZ = self._eq_solver.psi()
        # R = self._eq_solver.R
        # Z = self._eq_solver.Z
        # logger.debug(psiRZ,R,Z)
        # # Currents in the coils
        # self._machine.printCurrents()

        # # Forces on the coils
        # self._eq_solver.printForces()

        # print("\nSafety factor:\n\tpsi \t q")
        # for psi in [0.01, 0.9, 0.95]:
        #     print("\t{:.2f}\t{:.2f}".format(psi, self._eq_solver.q(psi)))

    def postprocess(self, current: FyEqAnalyze.TimeSlice):
        psi_norm = self.code.parameters.get("psi_norm", np.linspace(0, 1.0, 128))

        psi = psi_norm * (self._eq_solver.psi_bndry - self._eq_solver.psi_axis) + self._eq_solver.psi_axis

        # R0 = self._eq_solver.Rgeometric()

        # logger.debug(f"R0 = {R0},psi={psi}")
        R0 = current.vacuum_toroidal_field.r0
        B0 = self._eq_solver.fvac() / R0

        cache = {}
        cache["global_quantities"] = {"ip": self._eq_solver.plasmaCurrent()}

        cache["vacuum_toroidal_field"] = {"b0": B0, "r0": R0}

        cache["boundary"] = _not_found_

        # phi_bdry = self._eq_solver.tor_flux(self._eq_solver.psi_bndry)

        cache["profiles_1d"] = {
            "grid": {
                "psi_axis": self._eq_solver.psi_axis,
                "psi_boundary": self._eq_solver.psi_bndry,
                "psi_norm": psi_norm,
                # "rho_tor_boundary": np.sqrt(phi_bdry / (scipy.constants.pi * B0)),
                "rho_tor_norm": self._eq_solver.rhotor(psi),
            },
            "psi": psi,
            "psi_norm": psi_norm,
            "q": self._eq_solver.q(psi_norm),
            "pressure": self._eq_solver.pressure(psi_norm),
            "dpressure_dpsi": self._eq_solver.pprime(psi_norm),
            "f": self._eq_solver.fpol(psi_norm),
            "f_df_dpsi": self._eq_solver.ffprime(psi_norm),
        }

        psi2d = self._eq_solver.psi()

        trim = self.code.parameters.get("trim", 0)

        if trim > 0:
            grid2d = {
                "dim1": self._eq_solver.R_1D[trim:-trim],
                "dim2": self._eq_solver.Z_1D[trim:-trim],
            }
            psi2d = psi2d[trim:-trim, trim:-trim]
        else:
            grid2d = {
                "dim1": self._eq_solver.R_1D,
                "dim2": self._eq_solver.Z_1D,
            }

        cache["profiles_2d"] = {
            "type": "total",
            "grid_type": {"name": "rectangular", "index": 1},
            "grid": grid2d,
            "psi": psi2d,
        }
        current._cache = cache
        current._entry = None

        super().postprocess(current)


Equilibrium.register(["freegs"], EquilibriumFreeGS)



import os
import pathlib

from fytok.tokamak import Tokamak
from fytok.utils.logger import logger

from spdm.view.sp_view import display

WORKSPACE = "/home/salmon/workspace"  # "/ssd01/salmon_work/workspace/"

if __name__ == "__main__":
    output_path = pathlib.Path(f"{WORKSPACE}/output")

    tok = Tokamak(f"east+mdsplus://{WORKSPACE}/fytok_data/mdsplus/~t/?enable=efit_east", device="EAST", shot=70754)

    # f"file+GEQdsk://{WORKSPACE}/gacode/neo/tools/input/profile_data/g141459.03890", device="d3d"
    #     tok = Tokamak("file+geqdsk:///home/salmon/workspace/fytok_tutorial/tutorial/data/g900003.00230_ITER_15MA_eqdsk16HR.txt",
    #                   device='ITER', shot='900003', time=2.3)

    tok.refresh(time=5.0)
    
    prof2d = tok.equilibrium.time_slice.current.profiles_2d
    
    display(
        tok.equilibrium.time_slice.current.profiles_2d.psi,
        title=tok.title,
        styles={"interferometer": False},
        output=output_path / f"{tok.tag}_rz.svg",
    )

#     tok.advance()

#     display(tok,
#             title=tok.title,
#             styles={"interferometer": False},
#             output=output_path/f"{tok.tag}_rz.svg")
#     tok.advance()

#     display(tok,
#             title=tok.title,
#             styles={"interferometer": False},
#             output=output_path/f"{tok.tag}_rz.svg")



import pathlib

import numpy as np
from spdm.core.entry import Entry
from spdm.core.file import File
from spdm.utils.logger import logger


def mhdin_outline(r, z, w, h, a0, a1):
    return r, z, r+w, z+h


def sp_to_imas(data: dict):

    entry = Entry({})

    # entry["wall.description_2d[0].limiter.unit[0].outline.r"] = np.array(data["rsi"])
    # entry["wall.description_2d[0].limiter.unit[0].outline.z"] = np.array(data["zsi"])

    n_pf = 0
    if "rvs" in data:
        rvs = data['rvs']
        zvs = data['zvs']
        wvs = data['wvs']
        hvs = data['hvs']
        avs = data['avs']
        avs2 = data['avs2']

        for i, unit_name in enumerate(data['vsid']):
            a0 = -avs[i] * np.pi / 180.0
            a1 = -(avs2[i] * np.pi / 180.0 if avs2[i] != 0 else np.pi / 2)
            entry[f"wall.description_2d[0].vessel.unit[{i}].name"] = str(f"unit{int(unit_name)}")
            entry[f"wall.description_2d[0].vessel.unit[{i}].element[0].outline.r"] = np.array([
                rvs[i] - wvs[i] / 2.0 - hvs[i] / 2.0 * np.tan((np.pi / 2.0 + a1)),
                rvs[i] - wvs[i] / 2.0 + hvs[i] / 2.0 * np.tan((np.pi / 2.0 + a1)),
                rvs[i] + wvs[i] / 2.0 + hvs[i] / 2.0 * np.tan((np.pi / 2.0 + a1)),
                rvs[i] + wvs[i] / 2.0 - hvs[i] / 2.0 * np.tan((np.pi / 2.0 + a1)),
            ])

            entry[f"wall.description_2d[0].vessel.unit[{i}].element[0].outline.z"] = np.array([
                zvs[i] - hvs[i] / 2.0 - wvs[i] / 2.0 * np.tan(-a0),
                zvs[i] + hvs[i] / 2.0 - wvs[i] / 2.0 * np.tan(-a0),
                zvs[i] + hvs[i] / 2.0 + wvs[i] / 2.0 * np.tan(-a0),
                zvs[i] - hvs[i] / 2.0 + wvs[i] / 2.0 * np.tan(-a0),
            ])
            entry[f'wall.description_2d.0.vessel.unit.{i}.element.0.resistivity'] = data['rsisvs'][i]

    if "rf" in data:
        n_pf = len(data["rf"])
        for i in range(n_pf):
            # fmt:off
            pf_id=int(data["fcid"][i])-1
            entry[f"pf_active.coil[{i}].name"]                                  = f"PF{pf_id}"
            entry[f"pf_active.coil[{i}].identifier"]                            = f"PF{pf_id}"
            entry[f"pf_active.coil[{i}].element[0].geometry.geometry_type"]     = 2
            entry[f"pf_active.coil[{i}].element[0].geometry.rectangle.r"]       = float(data["rf"][i])
            entry[f"pf_active.coil[{i}].element[0].geometry.rectangle.z"]       = float(data["zf"][i])
            entry[f"pf_active.coil[{i}].element[0].geometry.rectangle.width"]   = float(data["wf"][i])
            entry[f"pf_active.coil[{i}].element[0].geometry.rectangle.height"]  = float(data["hf"][i])
            entry[f"pf_active.coil[{i}].element[0].turns_with_sign"]            = int(data['fcturn'][i])
            # fmt:on
    if "re" in data:
        re = data['re']
        ze = data['ze']
        we = data['we']
        he = data['he']
        ecturn = data['ecturn']
        elements_id = (np.array(data['ecid']) - 1).astype(int)
        for i in range(len(re)):
            c = elements_id[i]
            e = sum(elements_id[:i] == elements_id[i])
            entry[f'pf_active.coil[{i+n_pf}].name'] = f'OH{c}'
            entry[f'pf_active.coil[{i+n_pf}].identifier'] = f'OH{c}'
            entry[f'pf_active.coil[{i+n_pf}].element[0].name'] = f'OH{c}_{e}'
            entry[f'pf_active.coil[{i+n_pf}].element[0].identifier'] = f'OH{c}_{e}'
            entry[f'pf_active.coil[{i+n_pf}].element[0].turns_with_sign'] = ecturn[i]
            entry[f'pf_active.coil[{i+n_pf}].element[0].geometry.rectangle.r'] = re[i]
            entry[f'pf_active.coil[{i+n_pf}].element[0].geometry.rectangle.z'] = ze[i]
            entry[f'pf_active.coil[{i+n_pf}].element[0].geometry.rectangle.width'] = we[i]
            entry[f'pf_active.coil[{i+n_pf}].element[0].geometry.rectangle.height'] = he[i]
            entry[f'pf_active.coil[{i+n_pf}].element[0].geometry.geometry_type'] = 1

    if 'lpname' in data:
        n_flux = len(data['lpname'])
        for i in range(n_flux):
            # fmt:off
            flux_name                                                           = data['lpname'][i]
            entry[f'magnetics.flux_loop.{i}.name']                              = flux_name
            entry[f'magnetics.flux_loop.{i}.identifier']                        = flux_name
            entry[f'magnetics.flux_loop.{i}.position[0].r']                     = data["rsi"][i]
            entry[f'magnetics.flux_loop.{i}.position[0].z']                     = data["zsi"][i]
            entry[f'magnetics.flux_loop.{i}.type.index']                        = 1
            # fmt:on

    if 'mpnam2' in data:
        n_bp = len(data['mpnam2'])
        for i in range(n_bp):
            # fmt:off
            entry[f'magnetics.b_field_pol_probe.{i}.name']                      = data['mpnam2'][i]
            entry[f'magnetics.b_field_pol_probe.{i}.identifier']                = data['mpnam2'][i]
            entry[f'magnetics.b_field_pol_probe.{i}.position.r']                = data['xmp2'][i]
            entry[f'magnetics.b_field_pol_probe.{i}.position.z']                = data['ymp2'][i]
            entry[f'magnetics.b_field_pol_probe.{i}.length']                    = data['smp2'][i]
            entry[f'magnetics.b_field_pol_probe.{i}.poloidal_angle']            = -data['amp2'][i] / 180 * np.pi
            entry[f'magnetics.b_field_pol_probe.{i}.toroidal_angle']            = 0.0 / 180 * np.pi
            entry[f'magnetics.b_field_pol_probe.{i}.type.index']                = 1
            entry[f'magnetics.b_field_pol_probe.{i}.turns']                     = 1
            # fmt:on

    return entry


@File.register(["mhdin"])
class MHDINFile(File):
    """ READ mahchine description file (MHDIN)
        learn from omas
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self) -> Entry:
        if self.url.authority:
            raise NotImplementedError(f"{self.url}")

        data = File(self.url.path, mode="r", format="namelist").read().dump()

        return sp_to_imas(data["in3"])



# from __future__ import annotations
# @NOTE：
#   在插件中 from __future__ import annotations 会导致插件无法加载，
#   故障点是：typing.get_type_hints() 找不到类型， i.e. Code,TimeSeriesAoS

import typing
import numpy as np
import scipy.constants

from spdm.core.expression import Variable, Expression, Scalar, one, zero, derivative
from spdm.core.sp_property import sp_tree
from spdm.core.path import as_path
from spdm.utils.typing import array_type
from spdm.utils.tags import _not_found_


from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_sources import CoreSources
from fytok.modules.core_transport import CoreTransport
from fytok.modules.equilibrium import Equilibrium
from fytok.modules.transport_solver_numerics import TransportSolverNumerics, TransportSolverNumericsTimeSlice
from fytok.modules.utilities import *

from fytok.utils.atoms import atoms
from fytok.utils.logger import logger

from .bvp import solve_bvp

EPSILON = 1.0e-32


def derivative_(y: array_type, x: array_type, dc_index=None):
    res = derivative(y, x)

    if dc_index is not None:
        res[dc_index - 1 : dc_index + 1] = 0.5 * (res[dc_index - 1] + res[dc_index + 1])
        # res = np.zeros_like(x)
        # # res[:dc_index] = InterpolatedUnivariateSpline(x[:dc_index], y[:dc_index], ext=0).derivative()(x[:dc_index])
        # # res[dc_index:] = InterpolatedUnivariateSpline(x[dc_index:], y[dc_index:], ext=0).derivative()(x[dc_index:])
        # res[:dc_index] = (y[1 : dc_index + 1] - y[:dc_index]) / (x[1 : dc_index + 1] - x[:dc_index])
        # res[dc_index:-1] = (y[dc_index + 1 :] - y[dc_index:-1]) / (x[dc_index + 1 :] - x[dc_index:-1])
        # res[-1] = res[-2]

    return res


@sp_tree
class FyTrans(TransportSolverNumerics):
    r"""
    Solve transport equations $\rho=\sqrt{ \Phi/\pi B_{0}}$
    See  :cite:`hinton_theory_1976,coster_european_2010,pereverzev_astraautomated_1991`
    
        Solve transport equations :math:`\rho=\sqrt{ \Phi/\pi B_{0}}`
        See  :cite:`hinton_theory_1976,coster_european_2010,pereverzev_astraautomated_1991`

            Solve transport equations

            Current Equation

            Args:
                core_profiles       : profiles at :math:`t-1`
                equilibrium         : Equilibrium
                transports          : CoreTransport
                sources             : CoreSources
                boundary_condition  :

            Note:
                .. math ::  \sigma_{\parallel}\left(\frac{\partial}{\partial t}-\frac{\dot{B}_{0}}{2B_{0}}\frac{\partial}{\partial\rho} \right) \psi= \
                            \frac{F^{2}}{\mu_{0}B_{0}\rho}\frac{\partial}{\partial\rho}\left[\frac{V^{\prime}}{4\pi^{2}}\left\langle \left|\frac{\nabla\rho}{R}\right|^{2}\right\rangle \
                            \frac{1}{F}\frac{\partial\psi}{\partial\rho}\right]-\frac{V^{\prime}}{2\pi\rho}\left(j_{ni,exp}+j_{ni,imp}\psi\right)
                    :label: transport_current


                if :math:`\psi` is not solved, then

                ..  math ::  \psi =\int_{0}^{\rho}\frac{2\pi B_{0}}{q}\rho d\rho

            Particle Transport
            Note:

                .. math::
                    \left(\frac{\partial}{\partial t}-\frac{\dot{B}_{0}}{2B_{0}}\frac{\partial}{\partial\rho}\rho\right)\
                    \left(V^{\prime}n_{s}\right)+\frac{\partial}{\partial\rho}\Gamma_{s}=\
                    V^{\prime}\left(S_{s,exp}-S_{s,imp}\cdot n_{s}\right)
                    :label: particle_density_transport

                .. math::
                    \Gamma_{s}\equiv-D_{s}\cdot\frac{\partial n_{s}}{\partial\rho}+v_{s}^{pinch}\cdot n_{s}
                    :label: particle_density_gamma

            Heat transport equations

            Note:

                ion

                .. math:: \frac{3}{2}\left(\frac{\partial}{\partial t}-\frac{\dot{B}_{0}}{2B_{0}}\frac{\partial}{\partial\rho}\rho\right)\
                            \left(n_{i}T_{i}V^{\prime\frac{5}{3}}\right)+V^{\prime\frac{2}{3}}\frac{\partial}{\partial\rho}\left(q_{i}+T_{i}\gamma_{i}\right)=\
                            V^{\prime\frac{5}{3}}\left[Q_{i,exp}-Q_{i,imp}\cdot T_{i}+Q_{ei}+Q_{zi}+Q_{\gamma i}\right]
                    :label: transport_ion_temperature

                electron

                .. math:: \frac{3}{2}\left(\frac{\partial}{\partial t}-\frac{\dot{B}_{0}}{2B_{0}}\frac{\partial}{\partial\rho}\rho\right)\
                            \left(n_{e}T_{e}V^{\prime\frac{5}{3}}\right)+V^{\prime\frac{2}{3}}\frac{\partial}{\partial\rho}\left(q_{e}+T_{e}\gamma_{e}\right)=
                            V^{\prime\frac{5}{3}}\left[Q_{e,exp}-Q_{e,imp}\cdot T_{e}+Q_{ei}-Q_{\gamma i}\right]
                    :label: transport_electron_temperature
        """

    code: Code = {"name": "fy_trans", "copyright": "FyTok"}

    solver: str = "fy_trans_bvp_solver"

    primary_coordinate: str = "rho_tor_norm"

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        enable_momentum = self.code.parameters.enable_momentum or False
        enable_impurity = self.code.parameters.enable_impurity or False

        profiles_1d = self.profiles_1d

        ######################################################################################
        # 确定待求未知量

        unknowns = self.code.parameters.unknowns or list()

        # 极向磁通
        unknowns.append("psi_norm")

        # 电子
        # - 电子密度由准中性条件给出
        # - 电子温度，求解
        # - 电子转动，跟随离子，不求解
        unknowns.append("electrons/temperature")

        for s in self.ion_thermal:
            # 热化离子组份
            # - 密度 density,可求解，
            # - 温度 temperature,可求解，
            # - 环向转动 velocity/totoridal，可求解，

            unknowns.append(f"ion/{s}/density")
            unknowns.append(f"ion/{s}/temperature")
            if enable_momentum:
                unknowns.append(f"ion/{s}/velocity/toroidal")

        for s in self.ion_non_thermal:
            # 非热化离子组份，
            # - 密度 density ，可求解
            # - 温度 temperature, 无统一定义不求解，
            #    - He ash 温度与离子温度一致，alpha粒子满足慢化分布
            # - 环向转动 velocity/totoridal，无统一定义不求解
            #
            unknowns.append(f"ion/{s}/density")

        # if enable_impurity:
        #     # enable 杂质输运
        #     for s in self.impurities:
        #         unknowns.append(f"ion/{s}/density")
        #         unknowns.append(f"ion/{s}/temperature")
        #         if enable_momentum:
        #             unknowns.append(f"ion/{s}/velocity/toroidal")

        ######################################################################################
        # 声明主磁面坐标 primary_coordinate
        # 默认为 x=\bar{\rho}_{tor}=\sqrt{\frac{\Phi}{\Phi_{bdry}}}
        # \rho_{tor}= \sqrt{\frac{\Phi}{B_0 \pi}}

        x = Variable((i := 0), self.primary_coordinate)

        if self.primary_coordinate == "rho_tor_norm":
            x._metadata["label"] = r"\bar{\rho}_{tor}"

        ######################################################################################
        # 声明  variables 和 equations
        # profiles_1d = {self.primary_coordinate: x}
        # equations: typing.List[typing.Dict[str, typing.Any]] = []

        # 在 x=0 处边界条件唯一， flux=0 (n,T,u) or \farc{d \psi}{dx}=0 ( for psi )
        # 在 \rho_{bdry} 处边界条件类型可由参数指定
        bc_type = self.get_cache("boundary_condition_type", {})

        # 归一化/无量纲化单位
        # 在放入标准求解器前，系数矩阵需要无量纲、归一化
        units = self.code.parameters.units
        if units is _not_found_:
            units = {}

        else:
            units = units.__value__

        profiles_1d = self.profiles_1d

        profiles_1d[self.primary_coordinate] = x

        for s in unknowns:
            pth = Path(s)

            if pth[0] == "psi":
                label_p = r"\psi"
                label_f = r"\Psi"
                bc = bc_type.get(s, 1)

            if pth[0] == "psi_norm":
                label_p = r"\bar{\psi}"
                label_f = r"\bar{\Psi}"
                bc = bc_type.get(s, 1)

            if pth[-1] == "density":
                label_p = "n"
                label_f = r"\Gamma"
                bc = bc_type.get(s, None) or bc_type.get(f"*/density", 1)

            if pth[-1] == "temperature":
                label_p = "T"
                label_f = "H"
                bc = bc_type.get(s, None) or bc_type.get(f"*/temperature", 1)

            if pth[-1] == "toroidal":
                label_p = "u"
                label_f = r"\Phi"
                bc = bc_type.get(s, None) or bc_type.get(f"*/velocity/toroidal", 1)

            if pth[0] == "electrons":
                label_p += "_{e}"
                label_f += "_{e}"

            if pth[0] == "ion":
                label_p += f"_{{{pth[1]}}}"
                label_f += f"_{{{pth[1]}}}"

            profiles_1d[s] = Variable((i := i + 1), s, label=label_p)

            profiles_1d[f"{s}_flux"] = Variable((i := i + 1), f"{s}_flux", label=label_f)

            unit_profile = units.get(s, None) or units.get(f"*/{pth[-1]}", 1)

            unit_flux = units.get(f"{s}_flux", None) or units.get(f"*/{pth[-1]}_flux", 1)

            self.equations.append(
                {
                    "@name": s,
                    "identifier": s,
                    "units": (unit_profile, unit_flux),
                    "boundary_condition_type": bc,
                }
            )

        ###################################################################################################
        # 赋值属性
        # self.profiles_1d.update(profiles_1d)
        # self.equations = equations
        ##################################################################################################
        # 定义内部控制参数

        self._hyper_diff = self.code.parameters.hyper_diff or 0.001
        self._dc_pos = self.code.parameters.dc_pos or None

        # logger.debug([equ.identifier for equ in self.equations])

    def preprocess(self, *args, boundary_value=None, **kwargs) -> TransportSolverNumericsTimeSlice:
        """准备迭代求解
        - 方程 from self.equations
        - 初值 from initial_value
        - 边界值 from boundary_value
        """
        current: TransportSolverNumericsTimeSlice = super().preprocess(*args, **kwargs)

        profiles = self.profiles_1d

        rho_tor_norm = profiles.rho_tor_norm

        psi_norm = profiles.psi_norm

        eq0: Equilibrium.TimeSlice = self.inports["equilibrium/time_slice/0"].fetch()

        eq1: Equilibrium.TimeSlice = self.inports["equilibrium/time_slice/-1"].fetch()

        core0_1d: CoreProfiles.TimeSlice.Profiles1D = self.inports["core_profiles/time_slice/0/profiles_1d"].fetch()

        core1_1d: CoreProfiles.TimeSlice.Profiles1D = self.inports["core_profiles/time_slice/-1/profiles_1d"].fetch()

        ni = sum([ion.z * ion.get("density", zero) for ion in profiles.ion], zero)

        ni_flux = sum([ion.z * ion.get("density_flux", zero) for ion in profiles.ion], zero)

        # 杂质密度
        # for label in self.impurities:
        #     logger.debug(label)
        #     imp = core0_1d.ion[label]
        #     ni += imp.z * imp.density(rho_tor_norm)
        # ni_flux += core0_1d.ion[label].density_flux(rho_tor_norm)

        impurity_fraction = kwargs.get("impurity_fraction", 0.0)

        profiles.electrons["density"] = ni / (1.0 - impurity_fraction)

        profiles.electrons["density_flux"] = ni_flux / (1.0 - impurity_fraction)

        tranport: AoS[CoreTransport.Model.TimeSlice] = self.inports["core_transport/model/*"].fetch(profiles)

        sources: AoS[CoreSources.Source.TimeSlice] = self.inports["core_sources/source/*"].fetch(profiles)

        eq0_1d = eq0.profiles_1d

        # if psi_norm is _not_found_:
        #     # psi_norm = profiles.psi / (eq0_1d.grid.psi_boundary - eq0_1d.grid.psi_axis)
        #     psi_norm = Function(
        #         eq0_1d.grid.rho_tor_norm,
        #         eq0_1d.grid.psi_norm,
        #         name="psi_norm",
        #         label=r"\bar{\psi}",
        #     )(rho_tor_norm)

        # 设定全局参数
        # $R_0$ characteristic major radius of the device   [m]
        R0 = eq0.vacuum_toroidal_field.r0

        # $B_0$ magnetic field measured at $R_0$            [T]
        B0 = eq0.vacuum_toroidal_field.b0

        rho_tor_boundary = Scalar(current.grid.rho_tor_boundary)

        # $\frac{\partial V}{\partial\rho}$ V',             [m^2]
        vpr = eq0_1d.dvolume_drho_tor(psi_norm)

        # diamagnetic function,$F=R B_\phi$                 [T*m]
        fpol = eq0_1d.f(psi_norm)

        fpol2 = fpol**2

        # $q$ safety factor                                 [-]
        qsf = eq0_1d.q(psi_norm)

        gm1 = eq0_1d.gm1(psi_norm)  # <1/R^2>
        gm2 = eq0_1d.gm2(psi_norm)  # <|grad_rho_tor|^2/R^2>
        gm3 = eq0_1d.gm3(psi_norm)  # <|grad_rho_tor|^2>
        gm8 = eq0_1d.gm8(psi_norm)  # <R>

        if eq1 is _not_found_ or eq1 is None:
            one_over_dt = 0
            B1 = B0
            rho_tor_boundary_m = rho_tor_boundary
            vpr_m = vpr
            gm8_m = gm8
        else:
            dt = eq0.time - eq1.time

            if dt < 0:
                raise RuntimeError(f"dt={dt}<=0")
            elif np.isclose(dt, 0.0):
                one_over_dt = 0.0
            else:
                one_over_dt = one / dt

            B1 = eq1.vacuum_toroidal_field.b0
            rho_tor_boundary_m = eq1.profiles_1d.grid.rho_tor_boundary
            vpr_m = eq1.profiles_1d.dvolume_drho_tor(psi_norm)
            gm8_m = eq1.profiles_1d.gm8(psi_norm)

        k_B = (B0 - B1) / (B0 + B1) * one_over_dt

        k_rho_bdry = (rho_tor_boundary - rho_tor_boundary_m) / (rho_tor_boundary + rho_tor_boundary_m) * one_over_dt

        k_phi = k_B + k_rho_bdry

        rho_tor = rho_tor_boundary * rho_tor_norm

        inv_vpr23 = vpr ** (-2 / 3)

        k_vppr = 0  # (3 / 2) * k_rho_bdry - k_phi *　x * vpr(psi).dln()

        self._units = np.array(sum([equ.units for equ in self.equations], tuple()))

        X = current.grid.rho_tor_norm
        Y = np.zeros([len(self.equations) * 2, X.size])

        if boundary_value is None:
            boundary_value = {}

        if (initial_value := kwargs.get("initial_value", _not_found_)) is not _not_found_:
            for idx, equ in enumerate(self.equations):
                value = initial_value.get(equ.identifier, 0)
                Y[idx * 2] = value(X) if isinstance(value, Expression) else np.full_like(X, value)

        hyper_diff = self._hyper_diff

        for idx, equ in enumerate(self.equations):
            identifier = as_path(equ.identifier)
            path = identifier.parent
            quantity = identifier[-1]

            if quantity == "toroidal":
                quantity = "velocity/toroidal"
                path = path.parent

            bc_value = boundary_value.get(equ.identifier, Y[idx * 2][-1])

            match quantity:
                case "psi":
                    psi = identifier.get(profiles, zero)

                    psi_m = identifier.get(core1_1d, zero)(rho_tor_norm)

                    conductivity_parallel: Expression = zero

                    j_parallel: Expression = zero

                    for source in sources:
                        source_1d = source.profiles_1d
                        conductivity_parallel += as_path("conductivity_parallel").get(source_1d, zero)
                        j_parallel += as_path("j_parallel").get(source_1d, zero)

                    c = fpol2 / (scipy.constants.mu_0 * B0 * rho_tor * (rho_tor_boundary))

                    d_dt = one_over_dt * conductivity_parallel * (psi - psi_m) / c

                    D = vpr * gm2 / (fpol * rho_tor_boundary * 2.0 * scipy.constants.pi)

                    V = -k_phi * rho_tor_norm * conductivity_parallel

                    S = (
                        -vpr * (j_parallel) / (2.0 * scipy.constants.pi * rho_tor)
                        - k_phi
                        * conductivity_parallel
                        * (2 - 2 * rho_tor_norm * fpol.dln + rho_tor_norm * conductivity_parallel.dln)
                        * psi
                    ) / c

                    if bc_value is None:
                        bc_value = current.grid.psi_boundary

                    # at axis x=0 , dpsi_dx=0
                    bc = [[0, 1, 0]]

                    if bc_value is None:
                        assert equ.boundary_condition_type == 1
                        bc_value = current.grid.psi_boundary

                    # at boundary x=1
                    match equ.boundary_condition_type:
                        # poloidal flux;
                        case 1:
                            u = equ.units[1] / equ.units[0]
                            v = 0
                            w = bc_value * equ.units[1] / equ.units[0]

                        # ip, total current inside x=1
                        case 2:
                            Ip = bc_value
                            u = 0
                            v = 1
                            w = scipy.constants.mu_0 * Ip / fpol

                        # loop voltage;
                        case 3:
                            Uloop_bdry = bc_value
                            u = 0
                            v = 1
                            w = (dt * Uloop_bdry + psi_m) * (D - hyper_diff)

                        #  generic boundary condition y expressed as a1y'+a2y=a3.
                        case _:
                            if not isinstance(bc_value, (tuple, list)) or len(bc_value) != 3:
                                raise NotImplementedError(f"5: generic boundary condition y expressed as a1y'+a2y=a3.")
                            u, v, w = bc_value

                    bc += [[u, v, w]]

                case "psi_norm":
                    dpsi = current.grid.psi_boundary - current.grid.psi_axis

                    psi_norm = profiles.psi_norm

                    psi_norm_m = identifier.get(core1_1d, zero)(rho_tor_norm)

                    conductivity_parallel: Expression = zero

                    j_parallel: Expression = zero

                    for source in sources:
                        source_1d = source.profiles_1d
                        conductivity_parallel += as_path("conductivity_parallel").get(source_1d, zero)
                        j_parallel += as_path("j_parallel").get(source_1d, zero)

                    c = fpol2 / (scipy.constants.mu_0 * B0 * rho_tor * (rho_tor_boundary))

                    d_dt = one_over_dt * conductivity_parallel * (psi_norm - psi_norm_m) / c

                    D = vpr * gm2 / (fpol * rho_tor_boundary * 2.0 * scipy.constants.pi)

                    V = -k_phi * rho_tor_norm * conductivity_parallel

                    S = (
                        (
                            -vpr * (j_parallel) / (2.0 * scipy.constants.pi * rho_tor)
                            - k_phi
                            * conductivity_parallel
                            * (2 - 2 * rho_tor_norm * fpol.dln + rho_tor_norm * conductivity_parallel.dln)
                            * psi_norm
                        )
                        / c
                        / dpsi
                    )

                    if bc_value is None:
                        bc_value = current.grid.psi_norm[-1]

                    # at axis x=0 , dpsi_dx=0
                    bc = [[0, 1, 0]]

                    if bc_value is None:
                        assert equ.boundary_condition_type == 1
                        bc_value = current.grid.psi_boundary

                    # at boundary x=1
                    match equ.boundary_condition_type:
                        # poloidal flux;
                        case 1:
                            u = equ.units[1] / equ.units[0]
                            v = 0
                            w = bc_value * equ.units[1] / equ.units[0]

                        # ip, total current inside x=1
                        case 2:
                            Ip = bc_value
                            u = 0
                            v = 1
                            w = scipy.constants.mu_0 * Ip / fpol

                        # loop voltage;
                        case 3:
                            Uloop_bdry = bc_value
                            u = 0
                            v = 1
                            w = (dt * Uloop_bdry + psi_m) * (D - hyper_diff)

                        #  generic boundary condition y expressed as a1y'+a2y=a3.
                        case _:
                            if not isinstance(bc_value, (tuple, list)) or len(bc_value) != 3:
                                raise NotImplementedError(f"5: generic boundary condition y expressed as a1y'+a2y=a3.")
                            u, v, w = bc_value

                    bc += [[u, v, w]]

                case "density":
                    ns = (path / "density").get(profiles, zero)
                    ns_m = (path / "density").get(core1_1d, zero)(rho_tor_norm)

                    transp_D = zero
                    transp_V = zero

                    for transp in tranport:
                        transp_1d = transp.profiles_1d

                        transp_D += (path / "particles/d").get(transp_1d, zero)
                        transp_V += (path / "particles/v").get(transp_1d, zero)
                        # transp_F += (pth / "particles/flux").get(core_transp_1d, zero)

                    S = zero

                    for source in sources:
                        source_1d = source.profiles_1d
                        S += (path / "particles").get(source_1d, zero)

                    d_dt = one_over_dt * (vpr * ns - vpr_m * ns_m) * rho_tor_boundary

                    D = vpr * gm3 * transp_D / rho_tor_boundary  #

                    V = vpr * gm3 * (transp_V - rho_tor * k_phi)

                    S = vpr * (S - k_phi * ns) * rho_tor_boundary

                    # at axis x=0 , flux=0
                    bc = [[0, 1, 0]]

                    # at boundary x=1
                    match equ.boundary_condition_type:
                        case 1:  # 1: value of the field y;
                            u = equ.units[1] / equ.units[0]
                            v = 0
                            w = bc_value * equ.units[1] / equ.units[0]

                        case 2:  # 2: radial derivative of the field (-dy/drho_tor);
                            u = V
                            v = -1.0
                            w = bc_value * (D - hyper_diff)

                        case 3:  # 3: scale length of the field y/(-dy/drho_tor);
                            L = bc_value
                            u = V - (D - hyper_diff) / L
                            v = 1.0
                            w = 0
                        case 4:  # 4: flux;
                            u = 0
                            v = 1
                            w = bc_value
                        # 5: generic boundary condition y expressed as a1y'+a2y=a3.
                        case _:
                            if not isinstance(bc_value, (tuple, list)) or len(bc_value) != 3:
                                raise NotImplementedError(f"5: generic boundary condition y expressed as a1y'+a2y=a3.")
                            u, v, w = bc_value

                    bc += [[u, v, w]]

                case "temperature":
                    ns = (path / "density").get(profiles, zero)
                    Gs = (path / "density_flux").get(profiles, zero)
                    Ts = (path / "temperature").get(profiles, zero)

                    ns_m = (path / "density").get(core1_1d, zero)(profiles)
                    Ts_m = (path / "temperature").get(core1_1d, zero)(profiles)

                    energy_D = zero
                    energy_V = zero
                    energy_F = zero

                    flux_multiplier = zero

                    for transp in tranport:
                        transp_1d = transp.profiles_1d
                        flux_multiplier += transp_1d._parent.flux_multiplier

                        energy_D += (path / "energy/d").get(transp_1d, zero)
                        energy_V += (path / "energy/v").get(transp_1d, zero)
                        energy_F += (path / "energy/flux").get(transp_1d, zero)

                    if flux_multiplier is zero:
                        flux_multiplier = one

                    Q = zero

                    for source in sources:
                        source_1d = source.profiles_1d
                        Q += (path / "energy").get(source_1d, zero)

                    d_dt = (
                        one_over_dt
                        * (3 / 2)
                        * (vpr * ns * Ts - (vpr_m ** (5 / 3)) * ns_m * Ts_m * inv_vpr23)
                        * rho_tor_boundary
                    )

                    D = vpr * gm3 * ns * energy_D / rho_tor_boundary

                    V = vpr * gm3 * ns * energy_V + Gs * flux_multiplier - (3 / 2) * k_phi * vpr * rho_tor * ns

                    S = vpr * (Q - k_vppr * ns * Ts) * rho_tor_boundary

                    # at axis x=0, dH_dx=0
                    bc = [[0, 1, 0]]

                    # at boundary x=1
                    match equ.boundary_condition_type:
                        case 1:  # 1: value of the field y;
                            u = equ.units[1] / equ.units[0]
                            v = 0
                            w = bc_value * equ.units[1] / equ.units[0]

                        case 2:  # 2: radial derivative of the field (-dy/drho_tor);
                            u = V
                            v = -1.0
                            w = bc_value * (D - hyper_diff)

                        case 3:  # 3: scale length of the field y/(-dy/drho_tor);
                            L = bc_value
                            u = V - (D - hyper_diff) / L
                            v = 1.0
                            w = 0
                        case 4:  # 4: flux;
                            u = 0
                            v = 1
                            w = bc_value

                        case _:  # 5: generic boundary condition y expressed as a1y'+a2y=a3.
                            if not isinstance(bc_value, (tuple, list)) or len(bc_value) != 3:
                                raise NotImplementedError(f"5: generic boundary condition y expressed as a1y'+a2y=a3.")
                            u, v, w = bc_value

                    bc += [[u, v, w]]

                case "velocity/toroidal":
                    us = (path / "velocity/toroidal").get(profiles, zero)
                    ns = (path / "density").get(profiles, zero)
                    Gs = (path / "density_flux").get(profiles, zero)

                    us_m = (path / "velocity/toroidal").get(core1_1d, zero)(rho_tor_norm)
                    ns_m = (path / "density").get(core1_1d, zero)(rho_tor_norm)

                    chi_u = zero
                    V_u_pinch = zero

                    for transp in tranport:
                        transp_1d = transp.profiles_1d

                        chi_u += (path / "momentum/toroidal/d").get(transp_1d, zero)
                        V_u_pinch += (path / "momentum/toroidal/v").get(transp_1d, zero)

                    U = zero

                    for source in sources:
                        source_1d = source.profiles_1d
                        U += (identifier / "../../momentum/toroidal").get(source_1d, zero)

                    U *= gm8

                    ms = identifier.get(atoms).a

                    d_dt = one_over_dt * ms * (vpr * gm8 * ns * us - vpr_m * gm8_m * ns_m * us_m) * rho_tor_boundary

                    D = vpr * gm3 * gm8 * ms * ns * chi_u / rho_tor_boundary

                    V = (vpr * gm3 * ns * V_u_pinch + Gs - k_phi * vpr * rho_tor * ns) * gm8 * ms

                    S = vpr * (U - k_rho_bdry * ms * ns * us) * rho_tor_boundary

                    # at axis x=0, du_dx=0
                    bc = [[0, 1, 0]]

                    # at boundary x=1
                    match equ.boundary_condition_type:
                        case 1:  # 1: value of the field y;
                            u = equ.units[1]
                            v = 0
                            w = bc_value * equ.units[1]

                        case 2:  # 2: radial derivative of the field (-dy/drho_tor);
                            u = V
                            v = -1.0
                            w = bc_value * (D - hyper_diff)

                        case 3:  # 3: scale length of the field y/(-dy/drho_tor);
                            L = bc_value
                            u = V - (D - hyper_diff) / L
                            v = 1.0
                            w = 0
                        case 4:  # 4: flux;
                            u = 0
                            v = 1
                            w = bc_value

                        # 5: generic boundary condition y expressed as u y + v y'=w.
                        case _:
                            if not isinstance(bc_value, (tuple, list)) or len(bc_value) != 3:
                                raise NotImplementedError(f"5: generic boundary condition y expressed as a1y'+a2y=a3.")
                            u, v, w = bc_value

                    bc += [[u, v, w]]

                case _:
                    raise RuntimeError(f"Unknown equation of {equ.identifier}!")

            equ["coefficient"] = [d_dt, D, V, S]

            equ["boundary_condition_value"] = bc

        if (initial_value := kwargs.get("initial_value", _not_found_)) is not _not_found_:
            for idx, equ in enumerate(self.equations):
                d_dt, D, V, S = equ.coefficient
                y = Y[idx * 2]
                yp = derivative(y, X)
                Y[idx * 2 + 1] = -D(X, *Y) * yp + V(X, *Y) * y

        current.X = X
        current.Y = Y / self._units.reshape(-1, 1)

        return current

    def func(self, X: array_type, _Y: array_type, *args) -> array_type:
        dY = np.zeros([len(self.equations) * 2, X.size])

        hyper_diff = self._hyper_diff

        # 添加量纲和归一化系数，复原为物理量
        Y = _Y * self._units.reshape(-1, 1)

        for idx, equ in enumerate(self.equations):
            y = Y[idx * 2]
            flux = Y[idx * 2 + 1]

            _d_dt, _D, _V, _S = equ.coefficient

            try:
                d_dt = _d_dt(X, *Y, *args) if isinstance(_d_dt, Expression) else _d_dt
                D = _D(X, *Y, *args) if isinstance(_D, Expression) else _D
                V = _V(X, *Y, *args) if isinstance(_V, Expression) else _V
                S = _S(X, *Y, *args) if isinstance(_S, Expression) else _S
            except RuntimeError as error:
                raise RuntimeError(f"Error when calcuate {equ.identifier} {_S}") from error

            # yp = np.zeros_like(X)
            # yp[:-1] += 0.5 * ((y[1:] - y[:-1]) / (X[1:] - X[:-1]))  # derivative(flux, X)
            # yp[1:] += yp[:-1]
            # yp[0] = 0
            # yp[-1] *= 2
            yp = derivative(y, X)
            d_dr = (-flux + V * y + hyper_diff * yp) / (D + hyper_diff)

            # fluxp = np.zeros_like(X)
            # fluxp[:-1] = 0.5 * (flux[1:] - flux[:-1]) / (X[1:] - X[:-1])
            # fluxp[1:] += fluxp[:-1]
            # fluxp[0] = 0
            # flux[-1] *= 2

            fluxp = derivative(flux, X)
            dflux_dr = (S - d_dt + hyper_diff * fluxp) / (1.0 + hyper_diff)

            # if equ.identifier == "ion/alpha/density":
            #     dflux_dr[-1] = dflux_dr[-2]
            # if np.any(np.isnan(dflux_dr)):
            #     logger.exception(f"Error: {equ.identifier} nan in dflux_dr {_R._render_latex_()} {dflux_dr}")

            # 无量纲，归一化
            dY[idx * 2] = d_dr
            dY[idx * 2 + 1] = dflux_dr
            if equ.identifier in ["ion/alpha/density", "ion/He/density"]:
                #     dY[idx * 2, 0] = 0
                dY[idx * 2 + 1, -1] = 0

        dY /= self._units.reshape(-1, 1)

        return dY

    def bc(self, ya: array_type, yb: array_type, *args) -> array_type:
        x0, x1 = self.bc_pos

        bc = []

        ya = ya * self._units
        yb = yb * self._units
        for idx, equ in enumerate(self.equations):
            [u0, v0, w0], [u1, v1, w1] = equ.boundary_condition_value

            try:
                u0 = u0(x0, *ya, *args) if isinstance(u0, Expression) else u0
                v0 = v0(x0, *ya, *args) if isinstance(v0, Expression) else v0
                w0 = w0(x0, *ya, *args) if isinstance(w0, Expression) else w0
                u1 = u1(x1, *yb, *args) if isinstance(u1, Expression) else u1
                v1 = v1(x1, *yb, *args) if isinstance(v1, Expression) else v1
                w1 = w1(x1, *yb, *args) if isinstance(w1, Expression) else w1
            except Exception as error:
                logger.error(((u0, v0, w0), (u1, v1, w1)), exc_info=error)
                # raise RuntimeError(f"Boundary error of equation {equ.identifier}  ") from error

            y0 = ya[2 * idx]
            flux0 = ya[2 * idx + 1]

            y1 = yb[2 * idx]
            flux1 = yb[2 * idx + 1]

            # NOTE: 边界值量纲为 flux 通量，以 equ.units[1] 归一化
            bc.extend([(u0 * y0 + v0 * flux0 - w0) / equ.units[1], (u1 * y1 + v1 * flux1 - w1) / equ.units[1]])

        bc = np.array(bc)
        return bc

    def execute(
        self, current: TransportSolverNumericsTimeSlice, *previous: TransportSolverNumericsTimeSlice
    ) -> TransportSolverNumericsTimeSlice:
        current = super().execute(current, *previous)

        X = current.X
        Y = current.Y

        self.bc_pos = (X[0], X[-1])
        # 设定初值
        if Y is None:
            Y = np.zeros([len(self.equations) * 2, len(X)])

            for idx, equ in enumerate(self.equations):
                Y[2 * idx + 0] = (
                    equ.profile(X)
                    if isinstance(equ.profile, Expression)
                    else np.full_like(X, equ.profile if equ.profile is not _not_found_ else 0)
                )
                Y[2 * idx + 1] = (
                    equ.flux(X)
                    if isinstance(equ.flux, Expression)
                    else np.full_like(X, equ.flux if equ.flux is not _not_found_ else 0)
                )

        sol = solve_bvp(
            self.func,
            self.bc,
            X,
            Y,
            discontinuity=self.code.parameters.discontinuity or [],
            tol=self.code.parameters.tolerance or 1.0e-3,
            bc_tol=self.code.parameters.bc_tol or 1e6,
            max_nodes=self.code.parameters.max_nodes or 1000,
            verbose=self.code.parameters.verbose or 0,
        )

        current.X = sol.x
        current.Y = sol.y
        current.Yp = sol.yp
        current.rms_residuals = sol.rms_residuals

        logger.info(
            f"Solving the transport equation [{ 'success' if sol.success else 'failed'}]: {sol.message} , {sol.niter} iterations"
        )

        return current

    def postprocess(self, current: TransportSolverNumericsTimeSlice) -> TransportSolverNumericsTimeSlice:
        current = super().postprocess(current)

        X = current.X
        Y = current.Y * self._units.reshape(-1, 1)
        Yp = current.Yp * self._units.reshape(-1, 1)

        current["grid"] = CoreRadialGrid(
            {
                "rho_tor_norm": X,
                "psi_norm": Y[0],
                "psi_axis": current.grid.psi_axis,
                "psi_boundary": current.grid.psi_boundary,
                "rho_tor_boundary": current.grid.rho_tor_boundary,
            }
        )

        current["equations"] = []

        for idx, equ in enumerate(self.equations):
            d_dt, D, V, R = equ.coefficient

            current.equations.append(
                {
                    "@name": equ.identifier,
                    "boundary_condition_type": equ.boundary_condition_type,
                    "boundary_condition_value": equ.boundary_condition_value,
                    "profile": Y[2 * idx],
                    "flux": Y[2 * idx + 1],
                    "coefficient": [d_dt(X, *Y), D(X, *Y), V(X, *Y), R(X, *Y)],
                    "d_dr": Yp[2 * idx],
                    "dflux_dr": Yp[2 * idx + 1],
                }
            )

        return current


TransportSolverNumerics.register(["fy_trans"], FyTrans)



from __future__ import annotations
import typing
from spdm.core.path import update_tree
from spdm.core.entry import open_entry
from spdm.core.htree import HTree
from spdm.core.actor import Actor
from spdm.core.context import Context
from spdm.core.sp_property import sp_tree
from spdm.core.geo_object import GeoObject
from spdm.utils.tags import _not_found_

# ---------------------------------
from .utils.envs import *
from .utils.logger import logger

# ---------------------------------
from .modules.dataset_fair import DatasetFAIR
from .modules.summary import Summary
from .modules.core_profiles import CoreProfiles
from .modules.core_sources import CoreSources
from .modules.core_transport import CoreTransport
from .modules.ec_launchers import ECLaunchers
from .modules.equilibrium import Equilibrium
from .modules.ic_antennas import ICAntennas
from .modules.interferometer import Interferometer
from .modules.lh_antennas import LHAntennas
from .modules.magnetics import Magnetics
from .modules.nbi import NBI
from .modules.pellets import Pellets
from .modules.pf_active import PFActive
from .modules.tf import TF
from .modules.wall import Wall
from .modules.transport_solver_numerics import TransportSolverNumerics
from .modules.utilities import *

from .ontology import GLOBAL_ONTOLOGY
# from .modules.EdgeProfiles import EdgeProfiles
# from .modules.EdgeSources import EdgeSources
# from .modules.EdgeTransport import EdgeTransport
# from .modules.EdgeTransportSolver import EdgeTransportSolver
# ---------------------------------


@sp_tree
class Tokamak(Context):
    # fmt:off
    dataset_fair            : DatasetFAIR               

    # device
    wall                    : Wall                      

    # magnetics
    tf                      : TF                        
    pf_active               : PFActive                  
    magnetics               : Magnetics                 

    # aux
    ec_launchers            : ECLaunchers               
    ic_antennas             : ICAntennas                
    lh_antennas             : LHAntennas                
    nbi                     : NBI                       
    pellets                 : Pellets                   

    # diag
    interferometer          : Interferometer            

    # transport: state of device
    equilibrium             : Equilibrium               

    core_profiles           : CoreProfiles              
    core_transport          : CoreTransport             
    core_sources            : CoreSources               

    # edge_profiles         : EdgeProfiles              
    # edge_transport        : EdgeTransport             
    # edge_sources          : EdgeSources               
    # edge_transport_solver : EdgeTransportSolver       

    # solver
    transport_solver        : TransportSolverNumerics   

    summary                 : Summary                   
    # fmt:on

    @property
    def brief_summary(self) -> str:
        """综述模拟内容"""
        return f"""{FY_LOGO}
---------------------------------------------------------------------------------------------------
                                                Brief Summary
---------------------------------------------------------------------------------------------------
Dataset Description:
{self.dataset_fair}
---------------------------------------------------------------------------------------------------
Modules:
    transport_solver        : {self.transport_solver.code }
    equilibrium             : {self.equilibrium.code }

    core_profiles           : {self.core_profiles.code }             
    core_transport          : {', '.join([str(s.code).split(".")[-1] for s in self.core_transport.model])}
    core_sources            : {', '.join([str(s.code).split(".")[-1]  for s in self.core_sources.source])}
---------------------------------------------------------------------------------------------------
"""

    code: Code = {"name": "fy_tok"}

    @property
    def title(self) -> str:
        """标题，由初始化信息 dataset_fair.description"""
        return f"{self.dataset_fair.description}  time={self.time:.2f}s"

    @property
    def tag(self) -> str:
        """当前状态标签，由程序版本、用户名、时间戳等信息确定"""
        return f"{self.dataset_fair.description.tag}_{int(self.time*100):06d}"

    @property
    def shot(self) -> int:
        return self._shot

    @property
    def run(self) -> int:
        return self._run

    @property
    def device(self) -> str:
        return self._device

    def __init__(
        self,
        *args,
        device: str = _not_found_,
        shot: int = _not_found_,
        run: int = _not_found_,
        time: float = None,
        **kwargs,
    ):
        """
        用于集成子模块，以实现工作流。

        现有子模块包括： wall, tf, pf_active, magnetics, equilibrium, core_profiles, core_transport, core_sources, transport_solver

        :param args:   初始化数据，可以为 dict，str 或者  Entry。 输入会通过数据集成合并为单一的HTree，其子节点会作为子模块的初始化数据。
        :param device: 指定装置名称，例如， east，ITER, d3d 等
        :param shot:   指定实验炮号
        :param run:    指定模拟计算的序号
        :param time:   指定当前时间
        :param kwargs: 指定子模块的初始化数据，，会与args中指定的数据源子节点合并。
        """
        cache, entry, parent, kwargs = HTree._parser_args(*args, **kwargs)

        cache = update_tree(cache, kwargs)

        cache["dataset_fair"] = {"description": {"entry": entry, "device": device, "shot": shot or 0, "run": run or 0}}

        entry = open_entry(entry, shot=shot, run=run, local_schema=device, global_schema=GLOBAL_ONTOLOGY)

        super().__init__(cache, _entry=entry, _parent=parent)

        self._shot = shot
        self._run = run
        self._device = device
        self._metadata.setdefault("name", device)

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)

        self.core_profiles.initialize(time=self.time)
        self.equilibrium.initialize(time=self.time, pf_active=self.pf_active, wall=self.wall, magnetics=self.magnetics)
        self.core_sources.initialize(time=self.time, equilibrium=self.equilibrium, core_profiles=self.core_profiles)
        self.core_transport.initialize(time=self.time, equilibrium=self.equilibrium, core_profiles=self.core_profiles)
        self.transport_solver.initialize(
            time=self.time,
            equilibrium=self.equilibrium,
            core_profiles=self.core_profiles,
            core_sources=self.core_sources,
            core_transport=self.core_transport,
        )

    def refresh(self, *args, **kwargs) -> None:
        super().refresh(*args, **kwargs)

        self.core_profiles.refresh(time=self.time)
        self.equilibrium.refresh(time=self.time)
        self.core_sources.refresh(time=self.time)
        self.core_transport.refresh(time=self.time)

    def solve(self, *args, **kwargs) -> None:
        solver_1d = self.transport_solver.refresh(*args, time=self.time, **kwargs)
        profiles_1d = self.transport_solver.fetch()

        self.core_profiles.time_slice.current["profiles_1d"] = profiles_1d

        return solver_1d

    def flush(self):
        profiles_1d = self.transport_solver.fetch()

        self.core_profiles.time_slice.current["profiles_1d"] = profiles_1d

        self.core_profiles.flush()
        self.equilibrium.flush()
        self.core_transport.flush()
        self.core_sources.flush()
        self.transport_solver.flush()

        super().flush()

    def __view__(self, **kwargs) -> GeoObject | typing.Dict:
        geo = {}

        o_list = [
            "wall",
            "equilibrium",
            "pf_active",
            "magnetics",
            "interferometer",
            "tf",
            # "ec_launchers",
            # "ic_antennas",
            # "lh_antennas",
            # "nbi",
            # "pellets",
        ]

        for o_name in o_list:
            try:
                g = getattr(self, o_name, None)
                if g is None:
                    continue
                g = g.__view__(**kwargs)

            except Exception as error:
                logger.warning(f"Failed to get {g.__class__.__name__}.__view__ ! {error}")  # , exc_info=error
                # raise RuntimeError(f"Can not get {g.__class__.__name__}.__view__ !") from error
            else:
                geo[o_name] = g

        view_point = (kwargs.get("view_point", None) or "rz").lower()

        styles = {}

        if view_point == "rz":
            styles["xlabel"] = r"Major radius $R [m] $"
            styles["ylabel"] = r"Height $Z [m]$"

        styles["title"] = kwargs.pop("title", None) or self.title

        geo["$styles"] = styles

        return geo



from __future__ import annotations

from ..utils.logger import logger

from spdm.core.geo_object import GeoObject
from spdm.geometry.line import Line
from spdm.utils.tags import _not_found_
from ..ontology import interferometer


class Interferometer(interferometer._T_interferometer):
    def __view__(self, view_point="RZ", **kwargs) -> GeoObject:
        geo = {}
        match view_point.lower():
            case "rz":
                if self.channel is not _not_found_:
                    geo["channel"] = [
                        Line(
                            [channel.line_of_sight.first_point.r, channel.line_of_sight.first_point.z],
                            [channel.line_of_sight.second_point.r, channel.line_of_sight.second_point.z],
                            name=channel.name,
                            styles={"$matplotlib": {"color": "blue"}, "text": True},
                        )
                        for channel in self.channel
                    ]

        return geo





from  fytok._imas import _T_ids

class GasInjection(IDS):
    r"""Gas injection by a system of pipes and valves
        
        Note: GasInjection is an ids
    """
    _IDS="gas_injection"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)




import numpy as np
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_transport import CoreTransport
from fytok.modules.equilibrium import Equilibrium


class GyroBohm(CoreTransport.Model):
    """
    Heat conductivity Anomalous gyroBohm
    ===============================

    References:
    =============
    - Tokamaks, Third Edition, Chapter  4.16  ,p197,  J.A.Wesson 2003
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def update(self, *args, core_profiles: CoreProfiles = None, equilibrium: Equilibrium = None, **kwargs) -> float:
        residual = super().refresh(*args, core_profiles=core_profiles, equilibrium=equilibrium, **kwargs)

        prof = self.profiles_1d[-1]
        rho_tor_norm = core_profiles.profiles_1d.grid.rho_tor_norm
        psi_norm = core_profiles.profiles_1d.grid.psi_norm

        Te = np.asarray(core_profiles.profiles_1d.electrons.temperature) / 1.0e3
        ne = np.asarray(core_profiles.profiles_1d.electrons.density) / 1.0e19
        mu = 1.0 / np.asarray(equilibrium.profiles_1d.q(psi_norm))

        for ion in prof.ion:
            # ion.particles.d = 0
            # ion.particles.v = 0
            ion.energy.d = Chi_i
            ion.energy.v = 0

        # prof.electrons.particles.d = 0
        # prof.electrons.particles.v = 0
        # prof.electrons.energy.d = Chi_e
        # prof.electrons.energy.v = 0

        return residual


__SP_EXPORT__ = GyroBohm



from __future__ import annotations

from spdm.core.geo_object import GeoObject

from ..ontology import lh_antennas


class LHAntennas(lh_antennas._T_lh_antennas):
    def __view__(self, view_point="RZ", **kwargs) -> GeoObject:

        geo = {}
        styles = {}
        match view_point.lower():
            case "top":
                geo["antenna"] = [antenna.name for antenna in self.antenna]
                styles["antenna"] = {"$matplotlib": {"color": 'blue'}, "text": True}

        return geo 



import collections
import collections.abc
import functools
import typing
import numpy as np
import scipy.constants


from spdm.core.expression import Expression, Variable
from spdm.core.field import Field
from spdm.core.expression import Expression
from spdm.core.htree import List
from spdm.core.sp_property import sp_property, sp_tree
from spdm.core.time_series import TimeSeriesAoS
from spdm.geometry.curve import Curve
from spdm.core.geo_object import GeoObject, GeoObjectSet
from spdm.geometry.point import Point
from spdm.core.mesh import Mesh
from spdm.mesh.mesh_curvilinear import CurvilinearMesh
from spdm.numlib.contours import find_critical_points, find_contours

from spdm.utils.tags import _not_found_
from spdm.utils.typing import ArrayLike, NumericType, array_type, scalar_type, as_array

from fytok.modules.equilibrium import Equilibrium
from fytok.modules.utilities import *

from fytok.utils.logger import logger

PI = scipy.constants.pi

_R = Variable(0, "R")
_Z = Variable(1, "Z")

_T = typing.TypeVar("_T", float, array_type)


@sp_tree(coordiante1="psi_norm")
class MageticSurface(typing.Generic[_T]):
    @staticmethod
    def _shape_box(s: GeoObject):
        if isinstance(s, Point):
            r, z = s.points
            rmin = r
            rmax = r
            zmin = z
            zmax = z
            r_inboard = r
            r_outboard = r
            rzmin = r
            rzmax = r
        elif isinstance(s, Curve):
            r, z = s.points
            (rmin, zmin) = s.bbox.origin
            (rmax, zmax) = s.bbox.origin + s.bbox.dimensions
            rzmin = r[np.argmin(z)]
            rzmax = r[np.argmax(z)]

            # FIXME: 仅仅对当前磁面查找算法有效
            r_inboard = s.coordinates(0.5)[0]
            r_outboard = s.coordinates(0)[0]
        else:
            raise TypeError(f"Invalid type {type(s)}")
        return rmin, zmin, rmax, zmax, rzmin, rzmax, r_inboard, r_outboard

    def __init__(self, psi_norm: _T, curves: GeoObject | GeoObjectSet):
        self.psi_norm = psi_norm

        sbox = np.asarray([[*MageticSurface[_T]._shape_box(s)] for s in curves], dtype=float)
        (self.rmin, self.zmin, self.rmax, self.zmax, self.rzmin, self.rzmax, self.r_inboard, self.r_outboard) = sbox.T

    Rmin: _T
    Zmin: _T
    Rmax: _T
    Zmax: _T
    Rzmin: _T
    Rzmax: _T
    r_inboard: _T
    r_outboard: _T

    psi_norm: _T

    psi: _T = sp_property(units="Wb")

    phi: _T

    rho_tor: _T

    outline: Curve | GeoObjectSet

    x_point: GeoObjectSet = sp_property(alias="../coordinate_system/x_point")

    @sp_property
    def strike_point(self) -> GeoObjectSet:
        return

    @sp_property
    def active_limiter_point(self) -> GeoObjectSet:
        return NotImplemented

    @sp_property
    def geometric_axis(self) -> Point:
        return (self.Rmin + self.Rmax) * 0.5, (self.Zmin + self.Zmax) * 0.5

    @sp_property
    def minor_radius(self) -> _T:
        return (self.Rmax - self.Rmin) * 0.5

    @sp_property
    def elongation(self) -> _T:
        return (self.Zmax - self.Zmin) / (self.Rmax - self.Rmin)

    @sp_property
    def elongation_upper(self) -> _T:
        return (self.Zmax - (self.Zmax + self.Zmin) * 0.5) / (self.Rmax - self.Rmin)

    @sp_property
    def elongation_lower(self) -> _T:
        return ((self.Zmax + self.Zmin) * 0.5 - self.Zmin) / (self.Rmax - self.Rmin)

    @sp_property
    def triangularity(self) -> _T:
        return (self.Rzmax - self.Rzmin) / (self.Rmax - self.Rmin) * 2

    @sp_property
    def triangularity_upper(self) -> _T:
        return ((self.Rmax + self.Rmin) * 0.5 - self.Rzmax) / (self.Rmax - self.Rmin) * 2

    @sp_property
    def triangularity_lower(self) -> _T:
        return ((self.Rmax + self.Rmin) * 0.5 - self.Rzmin) / (self.Rmax - self.Rmin) * 2



import numpy as np
import scipy.constants
import typing
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.equilibrium import Equilibrium

from spdm.utils.typing import array_type
from spdm.utils.tags import _not_found_
from spdm.core.expression import Expression, Variable, zero
from spdm.core.sp_property import sp_tree

from fytok.modules.core_sources import CoreSources
from fytok.utils.atoms import atoms
from fytok.utils.logger import logger


@sp_tree
class Ohmic(CoreSources.Source):
    identifier = "collisional_equipartition"

    code = {"name": "ohmic", "description": "Fusion reaction"}  # type: ignore

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D) -> CoreSources.Source.TimeSlice:
        current: CoreSources.Source.TimeSlice = super().fetch(profiles_1d)

        #! +++ Radial electric field and ohmic heating:
        #      rho_loop13: DO irho=1,nrho
        #         if (sigma(irho).eq.0.0) then
        #           e_par(irho)=0.0
        #         else
        #           e_par(irho)              = ( curr_par(irho) - curr_ni_exp(irho)            &
        #                                   - curr_ni_imp(irho)*y(irho) ) / sigma(irho)
        #         end if
        #         qoh(irho)                = sigma(irho)*e_par(irho)**2 * control%ohmic_heating_multiplier
        #      END DO rho_loop13

        #    #!+++ Current diagnostics:
        #      fun7                       =  vpr * curr_par / 2.0e0_r8 / itm_pi * bt / fdia**2
        #      CALL integr2(nrho, rho, fun7, intfun7)
        #      curr_tot                   =  intfun7(nrho) * fdia(nrho)
        #      fun7                       =  vpr * (curr_ni_exp + curr_ni_imp * psi) / 2.0e0_r8 / itm_pi * bt / fdia**2
        #      CALL integr2(nrho, rho, fun7, intfun7)
        #      curr_ni                    =  intfun7(nrho) * fdia(nrho)

        #      fun7                       = qoh * vpr
        #      CALL integr2(nrho, rho, fun7, intfun7)
        #      qoh_tot                    = intfun7(nrho)



import numpy as np
from scipy import constants
from spdm.numlib.misc import array_like
from spdm.core.function import Function, function_like
from spdm.core.expression import derivative
from spdm.core.entry import _next_
from spdm.core.path import update_tree

from spdm.utils.tags import _not_found_

from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_transport import CoreTransport
from fytok.modules.equilibrium import Equilibrium
from fytok.utils.logger import logger


class NeoClassical(CoreTransport.Model):
    """
    Neoclassical Transport Model
    ===============================
    Neoclassical model, based on  Tokamaks, 3ed, J.A.Wesson 2003
    References:
    =============
    - Tokamaks, 3ed,  J.A.Wesson 2003
    """

    _metdata = {
        "identifier": "neoclassical",
        "code": {
            "name": "neoclassical",
            "description": f" Neoclassical model, based on  Tokamaks, 3ed, J.A.Wesson 2003",
        },
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def refresh(self, *args, equilibrium: Equilibrium, core_profiles: CoreProfiles, **kwargs) -> float:
        residual = super().refresh(*args, equilibrium=equilibrium, core_profiles=core_profiles, **kwargs)

        eV = constants.electron_volt
        B0 = equilibrium.vacuum_toroidal_field.b0
        R0 = equilibrium.vacuum_toroidal_field.r0

        core_profiles_1d = core_profiles.profiles_1d
        equilibrium_1d = equilibrium.profiles_1d

        rho_tor_norm = self.profiles_1d.grid_d.rho_tor_norm

        rho_tor_lcfs = self.profiles_1d.grid_d.rho_tor_boundary

        psi_norm = self.profiles_1d.grid_d.psi_norm

        q = equilibrium_1d.q(psi_norm)

        epsilon = rho_tor_norm * rho_tor_lcfs / R0
        epsilon12 = np.sqrt(epsilon)
        epsilon32 = epsilon ** (3 / 2)

        # Tavg = np.sum([ion.density*ion.temperature for ion in core_profile.ion]) / \
        #     np.sum([ion.density for ion in core_profile.ion])

        Te = core_profiles_1d.electrons.temperature(rho_tor_norm)
        Ne = core_profiles_1d.electrons.density(rho_tor_norm)
        Pe = core_profiles_1d.electrons.pressure(rho_tor_norm)
        dlnTe = derivative(core_profiles_1d.electrons.temperature, rho_tor_norm) / Te / rho_tor_lcfs
        dlnNe = derivative(core_profiles_1d.electrons.density, rho_tor_norm) / Ne / rho_tor_lcfs
        dlnPe = dlnNe + dlnTe

        vTe = np.sqrt(Te * eV / constants.electron_mass)

        # lnCoul = core_profiles_1d.coulomb_logarithm(rho_tor_norm)
        # Coulomb logarithm
        #  Ch.14.5 p727 Tokamaks 2003
        lnCoul = (14.9 - 0.5 * np.log(Ne / 1e20) + np.log(Te / 1000)) * (Te < 10) + (
            15.2 - 0.5 * np.log(Ne / 1e20) + np.log(Te / 1000)
        ) * (Te >= 10)

        # electron collision time , eq 14.6.1
        tau_e = 1.09e16 * ((Te / 1000) ** (3 / 2)) / Ne / lnCoul

        nu_star_e = R0 * q / vTe / tau_e / epsilon32

        # rho_tor[0] = max(1.07e-4*((Te[0]/1000)**(1/2))/np.abs(B0), rho_tor[0])  # Larmor radius,   eq 14.7.2

        # zeff = core_profiles_1d.zeff(rho_tor_norm)
        zeff = np.zeros_like(rho_tor_norm)
        n_i_total = np.zeros_like(rho_tor_norm)
        ###########################################################################################
        # Sec 14.11 Chang-Hinton formula for \Chi_i
        # Shafranov shift
        delta_ = derivative(equilibrium_1d.geometric_axis.r(psi_norm) - R0, rho_tor_norm) / rho_tor_lcfs

        # impurity ions
        nZ2_imp = np.sum(
            [
                (imp.z_ion_1d(rho_tor_norm) ** 2) * imp.density(rho_tor_norm)
                for imp in core_profiles_1d.ion
                if imp.is_impurity
            ]
        )

        f1 = (1.0 + (epsilon**2 + epsilon * delta_) * 3 / 2 + 3 / 8 * (epsilon**3) * delta_) / (
            1.0 + epsilon * delta_ / 2
        )
        f2 = (
            np.sqrt(1 - epsilon**2)
            * (1 + epsilon * delta_ / 2)
            / (1 + delta_ * (np.sqrt(1 - epsilon**2) - 1) / epsilon)
        )

        ###########################################################################################
        # Sec 14.12 Bootstrap current
        # Tokamak 3ed, 14.10

        # x = np.asarray(1.0 - (1-epsilon)**2/np.sqrt(1.0-epsilon**2)/(1+1.46*epsilon12))
        ft_e = 1.0 - (1 - epsilon) ** 2 / np.sqrt(1.0 - epsilon**2) / (1 + 1.46 * np.sqrt(epsilon))

        # fraction of trapped particle
        ft_i = np.sqrt(2 * epsilon)

        c1 = np.array(
            (4.0 + 2.6 * ft_e)
            / (1.0 + 1.02 * np.sqrt(nu_star_e) + 1.07 * nu_star_e)
            / (1.0 + 1.07 * epsilon32 * nu_star_e)
        )
        c3 = np.array(
            (7.0 + 6.5 * ft_e)
            / (1.0 + 0.57 * np.sqrt(nu_star_e) + 0.61 * nu_star_e)
            / (1.0 + 0.61 * epsilon32 * nu_star_e)
            - c1 * 5 / 2
        )

        j_bootstrap = np.asarray(c1 * dlnPe + c3 * dlnTe)

        ###########################################################################################
        # e_field_radial
        sum1 = 0.0
        sum2 = 0.0

        for ion in core_profiles_1d.ion:
            mi = ion.a
            Zi = ion.z_ion_1d(rho_tor_norm)
            Ti = ion.temperature(rho_tor_norm)
            Ni = ion.density(rho_tor_norm)

            # Larmor radius, Tokamaks 3ed, eq 14.7.2
            rho_i = 4.57e-3 * np.sqrt(mi * Ti / 1000) / B0

            # ion collision time Tokamaks 3ed, eq 14.6.2 p730
            tau_i = 6.6e17 * np.sqrt(mi) * ((Ti / 1000) ** (3 / 2)) / Ni / (1.1 * lnCoul)

            # thermal velocity
            v_Ti = np.sqrt(Ti * (eV / constants.m_p / mi))

            nZ2_ion = Ni * Zi * Zi
            zeff += nZ2_ion
            n_i_total += Zi * Ni
            #########################################################################
            # Sec 14.11 Chang-Hinton formula for \Chi_i
            if not ion.is_impurity:
                dlnTi = derivative(ion.temperature, rho_tor_norm) / Ti / rho_tor_lcfs
                dlnNi = derivative(ion.density, rho_tor_norm) / Ni / rho_tor_lcfs
                dlnPi = dlnNi + dlnTi
                alpha = nZ2_imp / nZ2_ion

                nu_star_i = R0 * q / epsilon32 / v_Ti / tau_i

                mu_star_i = nu_star_i * (1.0 + 1.54 * alpha)

                chi_i = (
                    0.66 * (1.0 + 1.54 * alpha) + (1.88 * np.sqrt(epsilon) - 1.54 * epsilon) * (1.0 + 3.75 * epsilon)
                ) / (1.0 + 1.03 * np.sqrt(mu_star_i) + 0.31 * mu_star_i)

                chi_i = chi_i * f1 + 0.59 * mu_star_i * epsilon / (1.0 + 0.74 * mu_star_i * epsilon32) * (
                    1.0 + 1.33 * alpha * (1.0 + 0.60 * alpha) / (1.0 + 1.79 * alpha)
                ) * (f1 - f2)

                chi_i = chi_i / epsilon32 * (q**2) * (rho_i**2) / (1.0 + 0.74 * mu_star_i * epsilon32)

                chi_i = array_like(rho_tor_norm, chi_i)

                self.profiles_1d.ion.put(
                    _next_,
                    {
                        "label": ion.label,
                        "a": ion.a,
                        "z": ion.z,
                        "is_impurity": ion.is_impurity,
                        "energy": {"d": function_like(chi_i, rho_tor_norm)},
                        "particles": {"d": function_like(chi_i / 3.0, rho_tor_norm)},
                    },
                )

                #########################################################################
                # for e_field_radial
                sum1 = sum1 + chi_i / 3.0 * derivative(ion.pressure, rho_tor_norm) / rho_tor_lcfs * Zi / Ti
                sum2 = sum2 + chi_i / 3.0 * nZ2_ion / Ti

            #########################################################################
            # Sec 14.12 Bootstrap current
            e3n2 = (epsilon**3) * (nu_star_i**2)
            c2 = c1 * Ti / Te
            c4 = (
                (
                    (-1.17 / (1.0 + 0.46 * ft_i) + 0.35 * np.sqrt(nu_star_i)) / (1 + 0.7 * np.sqrt(nu_star_i))
                    + 2.1 * e3n2
                )
                / (1 - e3n2)
                / (1 + e3n2)
                * c2
            )

            j_bootstrap += np.asarray(c2 * dlnPi + c4 * dlnTi)

        zeff /= n_i_total

        ###########################################################################################
        #
        self.profiles_1d["e_field_radial"] = sum1 / sum2

        ###########################################################################################
        #  Sec 14.10 Resistivity (spitzer)

        eta_s = 1.65e-9 * lnCoul * (Te / 1000) ** (-3 / 2)

        phi = ft_e / (1.0 + (0.58 + 0.20 * zeff) * nu_star_e)

        C = 0.56 / zeff * (3.0 - zeff) / (3.0 + zeff)

        eta = eta_s * zeff / (1 - phi) / (1.0 - C * phi) * (1.0 + 0.27 * (zeff - 1.0)) / (1.0 + 0.47 * (zeff - 1.0))

        self.profiles_1d["conductivity_parallel"] = function_like(1.0 / eta, rho_tor_norm)

        #########################################################################
        #  Sec 14.12 Bootstrap current

        fpol = equilibrium_1d.fpol(psi_norm)
        j_bootstrap = array_like(
            rho_tor_norm,
            -j_bootstrap
            * ft_e
            / (2.4 + 5.4 * ft_e + 2.6 * ft_e**2)
            * Pe
            * fpol
            * q
            / rho_tor_norm
            / rho_tor_lcfs**2
            / (2.0 * constants.pi * B0),
        )

        self.profiles_1d["j_bootstrap"] = function_like(j_bootstrap, rho_tor_norm)
        self.profiles_1d["j_ohmic"] = function_like(core_profiles_1d.e_field.parallel(rho_tor_norm) / eta, rho_tor_norm)

        return residual


__SP_EXPORT__ = NeoClassical



import typing
import scipy.constants
from spdm.core.expression import Variable, Expression, zero
from spdm.core.sp_property import sp_tree



from fytok.modules.equilibrium import Equilibrium
from fytok.modules.core_sources import CoreSources
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.utilities import *

from fytok.utils.atoms import atoms

from fytok.utils.logger import logger

PI = scipy.constants.pi


@sp_tree
class SynchrotronRadiation(CoreSources.Source):
    identifier = "synchrotron"

    code = {
        "name": "synchrotron_radiation",
        "description": """
    Source from Synchrotron radition 
    
    """,
    }  # type: ignore

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D) -> CoreSources.Source.TimeSlice:
        current: CoreSources.Source.TimeSlice = super().fetch(profiles_1d)

        source_1d = current.profiles_1d

        ne = profiles_1d.electrons.density
        Te = profiles_1d.electrons.temperature

        eq_1d: Equilibrium.TimeSlice.Profiles1D = self.inports["equilibrium/time_slice/current/profiles_1d"].fetch()

        B0 = np.abs(eq_1d._parent.vacuum_toroidal_field.b0)
        R0 = eq_1d._parent.vacuum_toroidal_field.r0

        x = profiles_1d.rho_tor_norm

        if True:
            #   Reference: (GACODE)
            #    Synchrotron synchrotron
            #        - Trubnikov, JETP Lett. 16 (1972) 25.

            psi_norm = Function(eq_1d.grid.rho_tor_norm, eq_1d.grid.psi_norm, label=r"\bar{\psi}")(x)

            r_min = eq_1d.major_radius(psi_norm)

            aspect_rat = R0 / r_min

            r_coeff = 0.8  # Reflection coefficient (Rosenbluth)

            me = scipy.constants.electron_mass
            e = scipy.constants.elementary_charge
            PI = scipy.constants.pi
            c = scipy.constants.speed_of_light

            wpe = np.sqrt(4 * PI * ne * e**2 / me)
            wce = e * B0 / (me * c)
            g = Te / scipy.constants.electron_volt / (me * c**2)
            phi = (
                60
                * g**1.5
                * np.sqrt((1.0 - r_coeff) * (1 + 1 / aspect_rat / np.sqrt(g)) / (r_min * wpe**2 / c / wce))
            )

            qsync = me / (3 * PI * c) * g * (wpe * wce) ** 2 * phi

        else:
            qsync = 6.2e-22 * B0**2.0 * ne * Te

        source_1d.electrons.energy -= qsync

        return current


CoreSources.Source.register(["synchrotron_radiation"], SynchrotronRadiation)



import pathlib
import itertools
import re
import numpy as np
from scipy.interpolate import CubicHermiteSpline

from spdm.utils.logger import logger
from spdm.core.file import File
from spdm.core.entry import Entry
from spdm.core.expression import Expression


def sp_read_pfile_txt(path: str | pathlib.Path):
    """
    :param file: input file / file path
    :return: profile object
    """
    s = "ne(10^20/m^3)"
    match = re.match(r"(\w+)\((.*)\)", s)

    res = {}
    with open("/home/salmon/workspace/fytok_ext/examples/data/Pfile_H_6_2", mode="r") as fid:
        while True:
            head = fid.readline()
            if head == "":
                break
            line_num, psi_norm, profile_name, dprofile_dpsi = head.split()
            line_num = int(line_num)
            match = re.match(r"(\w+)\((.*)\)", profile_name)
            if match:
                profile_name = match.group(1)
                unit = match.group(2)
            else:
                unit = "-"

            data = np.asarray([[float(v) for v in x.split()] for x in itertools.islice(fid, line_num)])
            ppoly = CubicHermiteSpline(data[:, 0], data[:, 1], data[:, 2])
            res[profile_name] = Expression(ppoly, domain=data[:, 0], name=profile_name, unit=unit)
    return res


def sp_to_imas(data: dict):
    entry = Entry({})

    core_profiles_1d = entry.child("core_profiles/time_slice/0/profiles_1d/")

    core_profiles_1d["electrons/temperature"] = data["te"]
    core_profiles_1d["electrons/density"] = data["ne"]
    core_profiles_1d["t_i_average"] = data["ti"]
    core_profiles_1d["n_i_thermal_total"] = data["ni"]
    core_profiles_1d["pressure_parallel"] = data["pb"]
    core_profiles_1d["pressure_ion_total"] = data["ptot"]
    # fmt:on

    return entry


@File.register(["pfile"])
class PFile(File):
    """Read pfile file  (from gacode)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self) -> Entry:
        if self.url.authority:
            raise NotImplementedError(f"{self.url}")

        path = pathlib.Path(self.url.path)

        if path.suffix.lower() in [".nc", ".h5"]:
            data = File(path, mode="r").read().dump()
        else:
            data = sp_read_pfile_txt(path)

        return sp_to_imas(data)

    def write(self, d, *args, **kwargs):
        raise NotImplementedError(f"TODO: write ITERDB {self.url}")



from __future__ import annotations

import numpy as np
from spdm.core.geo_object import GeoObject
from spdm.geometry.polygon import Polygon
from spdm.utils.constants import TWOPI
from spdm.utils.typing import _not_found_
from ..utils.logger import logger
from ..ontology import tf


class TF(tf._T_tf):

    def __view__(self, view_point="RZ", **kwargs) -> GeoObject:
        geo = {}
        styles = {}
        r0 = self.r0

        match view_point.lower():
            case "rz_":
                conductor = self.coil[0].conductor[0]
                geo["coils"] = Polygon(conductor.elements.start_points.r,
                                       conductor.elements.start_points.z,
                                       name=self.coil[0].name)

            case "top":
                if self.is_periodic == 0:
                    coils_n = self.coils_n
                    d_phi = TWOPI/self.coils_n

                    cross_section = self.coil[0].conductor[0].cross_section
                    r = cross_section.delta_r
                    phi = cross_section.delta_phi
                    name = self.coil[0].name
                    geo["coils"] = [
                        Polygon((r0+r)*np.cos(phi+d_phi*i),
                                (r0+r)*np.sin(phi+d_phi*i),
                                name=name+f"{i}") for i in range(coils_n)]

                else:
                    geo["coils"] = [
                        Polygon(
                            (r0 + coil.conductor[0].cross_section.delta_r) *
                            np.cos(coil.conductor[0].cross_section.delta_phi),
                            (r0 + coil.conductor[0].cross_section.delta_r) *
                            np.sin(coil.conductor[0].cross_section.delta_phi),
                            name=coil.name) for coil in self.coil]

        return geo




import pathlib
import re

import numpy as np
import pandas as pd
from scipy import constants
from spdm.core.expression import Piecewise, Variable
from spdm.core.File import File
from spdm.core.path import update_tree
from spdm.numlib.smooth import smooth_1d
from fytok.utils.logger import logger

TWOPI = 2.0*constants.pi


def load_core_profiles(profiles, grid):

    bs_r_norm = profiles["x"].values

    # Core profile
    r_ped = 0.96  # np.sqrt(0.88)
    i_ped = np.argmin(np.abs(bs_r_norm-r_ped))
    # fmt:off
    bs_psi_norm = profiles["Fp"].values
    # bs_psi = bs_psi_norm*(psi_boundary-psi_axis)+psi_axis

    b_Te =    smooth_1d(profiles["TE"].values,     bs_r_norm, i_end=i_ped-10, window_len=21)*1000
    b_Ti =    smooth_1d(profiles["TI"].values,     bs_r_norm, i_end=i_ped-10, window_len=21)*1000
    b_ne =    smooth_1d(profiles["NE"].values,     bs_r_norm, i_end=i_ped-10, window_len=21)*1.0e19
    b_nDT =   smooth_1d(profiles["Nd+t"].values,   bs_r_norm, i_end=i_ped-10, window_len=21)*1.0e19*0.5
    b_nHe =   smooth_1d(profiles["Nath"].values,   bs_r_norm, i_end=i_ped-10, window_len=21)*1.0e19
    b_nImp =  smooth_1d(profiles["Nz"].values,     bs_r_norm, i_end=i_ped-10, window_len=21)*1.0e19
    b_zeff = profiles["Zeff"].values
    # fmt:on

    z_eff_star = b_zeff-(b_nDT*2.0+4*b_nHe)/b_ne
    z_imp = 1-(b_nDT*2.0+2*b_nHe)/b_ne
    b = -2*z_imp/(0.02+0.0012)
    c = (z_imp**2-0.02*z_eff_star)/0.0012/(0.02+0.0012)

    z_Ar = np.asarray((-b+np.sqrt(b**2-4*c))/2)
    z_Be = np.asarray((z_imp-0.0012*z_Ar)/0.02)
    # b_nDT = b_ne * (1.0 - 0.02*4 - 0.0012*18) - b_nHe*2.0

    # Zeff = Function(bs_r_norm, baseline["Zeff"].values)
    # e_parallel = baseline["U"].values / (TWOPI * R0)

    return {
        "time": 0.0,
        "grid": grid,
        "electrons": {"label": "e", "density_thermal":  b_ne,   "temperature": b_Te, },
        "ion": [
            {"label": "D",  "density_thermal":      b_nDT,      "temperature": b_Ti},
            {"label": "T",  "density_thermal":      b_nDT,      "temperature": b_Ti},
            {"label": "He", "density_thermal":      b_nHe,      "temperature": b_Ti,  "density_fast": True},
            {"label": "Be", "density_thermal":  0.02*b_ne,      "temperature": b_Ti, "z_ion_1d": z_Be,  "is_impurity": True},
            {"label": "Ar", "density_thermal": 0.0012*b_ne,     "temperature": b_Ti, "z_ion_1d": z_Ar,  "is_impurity": True},
        ],
        # "e_field": {"parallel":  Function(e_parallel,bs_r_norm)},
        # "conductivity_parallel": Function(baseline["Joh"].values*1.0e6 / baseline["U"].values * (TWOPI * grid.r0),bs_r_norm),

        "rho_tor":          profiles["rho"].values,
        "zeff":             profiles["Zeff"].values,
        "vloop":            profiles["U"].values,
        "j_ohmic":          profiles["Joh"].values*1.0e6,
        "j_non_inductive":  profiles["Jnoh"].values*1.0e6,
        "j_bootstrap":      profiles["Jbs"].values*1.0e6,
        "j_total":          profiles["Jtot"].values*1.0e6,
        "XiNC":             profiles["XiNC"].values,

        "ffprime":          profiles["EQFF"].values*1.0e6,
        "pprime":           profiles["EQPF"].values*1.0e6,
    }


def load_core_transport(profiles, grid, R0: float, B0: float = None):

    bs_r_norm = profiles["x"].values
    bs_psi_norm = profiles["Fp"].values

    _x = Variable(0, "rho_tor_norm")

    # Core profiles
    r_ped = 0.96  # np.sqrt(0.88)
    i_ped = np.argmin(np.abs(bs_r_norm-r_ped))

    # Core Transport

    Cped = 0.17
    Ccore = 0.4
    # Function( profiles["Xi"].values,bs_r_norm)  Cped = 0.2
    chi = Piecewise([Ccore*(1.0 + 3*(_x**2)),   Cped],        [(_x < r_ped), (_x >= r_ped)])
    chi_e = Piecewise([0.5 * Ccore*(1.0 + 3*(_x**2)),   Cped],  [(_x < r_ped), (_x >= r_ped)])

    D = 0.1*(chi+chi_e)

    v_pinch_ne = -0.6 * D * _x / R0
    v_pinch_Te = 2.5 * chi_e * _x / R0
    v_pinch_ni = D * _x / R0
    v_pinch_Ti = chi * _x / R0

    return {
        "grid_d": grid,
        "conductivity_parallel":  profiles["Joh"].values*1.0e6 / profiles["U"].values * (TWOPI * R0),
        "electrons": {
            "label": "e",
            "particles":   {"d": D,     "v": v_pinch_ne},
            "energy":      {"d": chi_e, "v": v_pinch_Te},
        },
        "ion": [
            {
                "label": "D",
                "particles": {"d":  D, "v": v_pinch_ni},
                "energy": {"d":  chi, "v": v_pinch_Ti},
            },
            {
                "label": "T",
                "particles": {"d":  D, "v": v_pinch_ni},
                "energy": {"d":  chi, "v": v_pinch_Ti},
            },
            {
                "label": "He",
                "particles": {"d": D, "v": v_pinch_ni},
                "energy": {"d": chi, "v": v_pinch_Ti},
            }
        ]
    }


def load_core_source(profiles, grid, R0: float, B0: float = None):
    bs_r_norm = profiles["x"].values
    bs_psi_norm = profiles["Fp"].values

    _x = Variable(0, "rho_tor_norm")

    S = 9e20 * np.exp(15.0*(_x**2-1.0))

    Q_e = (profiles["Poh"].values
           + profiles["Pdte"].values
           + profiles["Paux"].values
           - profiles["Peic"].values
           - profiles["Prad"].values
           # - profiles["Pneu"].values
           )*1e6/constants.electron_volt

    Q_DT = (profiles["Peic"].values
            + profiles["Pdti"].values
            + profiles["Pibm"].values
            )*1e6/constants.electron_volt

    Q_He = (- profiles["Pdti"].values
            - profiles["Pdte"].values
            )*1e6/constants.electron_volt

    # Core Source
    return {

        "grid": grid,
        "j_parallel": (
            # profiles["Jtot"].values
            profiles["Joh"].values
            # + profiles["Jbs"].values
            + profiles["Jnb"].values
            + profiles["Jrf"].values
        ) * 1e6,  # A/m^2
        "electrons": {"label": "e",  "particles": S, "energy": Q_e},
        "ion": [
            {"label": "D",          "particles": S*0.5,      "energy": Q_DT*0.5},
            {"label": "T",          "particles": S*0.5,      "energy": Q_DT*0.5},
            {"label": "He",         "particles": S*0.01,     "energy": Q_He}
        ]}


def load_scenario_ITER(path):

    path = pathlib.Path(path)

    # eq_file = path/"Standard domain R-Z/Medium resolution - 129x257/g900003.00230_ITER_15MA_eqdsk16MR.txt"

    profiles_file = next(path.glob("*.xls"))

    logger.info(f"Load scenario/profiles from {profiles_file}")

    excel_file = pd.read_excel(profiles_file, sheet_name=1)

    scenario = {
        "name": "15MA Inductive at burn-ASTRA",
        "description": f"{profiles_file.name}"
    }

    desc = {}
    for s in excel_file.iloc[0, 3:7]:
        res = re.match(r'(\w+)=(\d+\.?\d*)(\D+)', s)
        desc[res.group(1)] = (float(res.group(2)), str(res.group(3)))

    time = 0.0

    vacuum_toroidal_field = {"r0": desc["R"][0], "b0": desc["B"][0]}

    profiles = pd.read_excel(profiles_file, sheet_name=1, header=10, usecols="B:BN")
    bs_r_norm = profiles["x"].values
    bs_psi_norm = profiles["Fp"].values

    grid = {
        "rho_tor_norm":  bs_r_norm,
        "rho_tor":  profiles["rho"].values,
        "psi_norm": bs_psi_norm,
        # "psi_boundary": psi_boundary,
        # "psi_axis": psi_axis,
    },

    scenario["core_profiles"] = {"time_slice": [{
        "time": time,
        'vacuum_toroidal_field': vacuum_toroidal_field,
        "profiles_1d": load_core_profiles(profiles, grid=grid),
    }]}

    scenario["core_transport"] = {"model": [{
        "code": {"name": "dummy"},
        "time_slice": [{
            "time": time,
            'vacuum_toroidal_field': vacuum_toroidal_field,
            "profiles_1d": load_core_transport(profiles, vacuum_toroidal_field["r0"]),
        }]
    }]}

    scenario["core_sources"] = {"source": [{
        "code": {"name": "dummy"},
        "time_slice": [{
            "time": time,
            'vacuum_toroidal_field': vacuum_toroidal_field,
            "profiles_1d": load_core_source(profiles, vacuum_toroidal_field["r0"])
        }]
    }]}

    # /"Increased domain R-Z/Medium resolution - 129x257/g900003.00230_ITER_15MA_eqdsk16VVMR.txt"
    eq_file = next(path.glob("**/*.txt"))

    logger.info(f"Load scenario/equilibrium from {eq_file}")

    update_tree(scenario, None, File(eq_file, format="GEQdsk").read().dump())

    return scenario


def load_scenario(path):
    return load_scenario_ITER(path)
